/*
 ============================================================================
 Name        : global.h
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */
/****************************************************************
*
*                 global.h (global include file for Drodi project)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

#ifndef GLOBAL_H_
#define GLOBAL_H_
/* Include for platform independent modules */

/*--- Standard Includes ----------------------------------------*/
#include <stddef.h>                     /* Standard definitions         */
#include <string.h>                     /* strcpy/memset                */
#include <ctype.h>                      /* isdigit/toupper...           */
#include <stdio.h>                      /* sprintf/fopen...             */
#include <stdlib.h>                     /* malloc/free...               */
#include <time.h>                       /* clock(),...                  */
#include <sys/time.h>                   /* gettimeofday()               */
#include <limits.h>                     /* CHAR_BIT                     */
#include <stdint.h>                     /* intxx_t data types           */
#include <locale.h>                     /* setlocale()                  */
#include <math.h>                       /* sqrt()                       */
#include <signal.h>                     /* signal(),...                 */
#include <errno.h>                      /* strerror()                   */
#include <unistd.h>                     /* getpid(), sysconf()...       */
#include <sys/types.h>                  /* getpid()                     */
#include <sys/resource.h>               /* getrlimit(), setrlimit()     */
#include <sys/mman.h>                   /* mmap()                       */
#include <semaphore.h>                  /* sem_xxx()                    */
#include <sys/wait.h>                   /* wait()                       */
#include <malloc.h>                     /* mallinfo2()                  */

/* Define's for compilation time code generation control. */
//#define DEBUGMEMORY                         /* Generate debug code for memory problems detected by Valgrind.*/
                                            /* See myfree() comments for information. */
//#define DEBUGCODE                           /* Generate additional debug code. */
//#define VERBOSE                             /* Display inferences if DEBUGCODE is defined and only one core enabled. */
//#define DEBUGTREE                             /* Generate debug code for discrimination tree integrity.*/
//#define DEBUGTREE2                            /* Generate debug code for discrimination tree queries. The checks will */
                                              /* only be done if DEBUGTREE is not defined. See ChecDscTQueries() */
                                              /* function global comments for additional information. */
#ifdef DEBUGTREE2
#define NBOFTESTS 3                           /* Number of literals to be tested by ChecDscTQueries() function. */
#endif
//#define SATDEBUGCODE1                       /* Generate additional debug code specific to check coherence */
                                            /* of hashchcmp elements ordered sub-queue. */
//#define SATDEBUGCODE2                       /* Generate additional debug code specific to check coherence */
                                            /* of satclause elements ordered sub-queue. */
//#define VERIFYSATPROOF                      /* Verify that the SAT part of the proof is coherent. */
#define DEBUGKBDATA 0                       /* Set to 0 for normal operations. Set to 1 to force print KB data even in */
                                            /* TIMEOUT or NOMEMORY conditions. */
//#define MURMUR3                             /* Use Murmur3 hash function (default is to use KRM hash function). */
//#define BENCHMARKING                        /* Include code for benchmarking summary output. */
                                            /* VERY IMPORTANT: If BENCHMARKING is defined then the BENCHDIMENSION parameter */
                                            /* below must be adjusted appropriately. */
#ifdef BENCHMARKING
#define BENCHTYPE 3                         /* 1 -> report solving times for problems solved. */
                                            /* 2 -> report THM, SAT, CSA, CAX or UNS for problems solved. */
                                            /* 3 -> report 1 and 2 above for problems solved. */
#define VERIFYSATFINALSTATUS 0              /* If this equals one then SAT final status is verified after a problem solving attempt. */
                                            /* Only if BENCHMARKING is defined this can be set to one, otherwise it must be set */
                                            /* to zero. See VerifySatStatus() function for details. */
#define BENCHDIMENSION 381                  /* Adjust this to the number of problems per run. */
#else
#define VERIFYSATFINALSTATUS 0              /* Don't modify this. See comments above. */
#endif
//#define MEMORYTUNNING                       /* If this is defined some code is generated for memory limit fine tuning and the */
                                            /* SymMalloc() function is defined and called from Initialize_pgm() and Clean4Exit() */
                                            /* functions to simulate huge memory allocation. */
#ifdef MEMORYTUNNING
#define ADDTLMEMORY 4000000000              /* Additional memory to be allocated if MEMORYTUNNING is defined. */
#endif
#define ALLOWDOTSINFNAME                    /* If this is defined then dots are allowed in "name" field and possibly in other */
                                            /* positions inside the of a TPTP annotated formula. */
#define ADDSTRAT2SEED                       /* If this is defined then when shuffling is active if the active strategy is odd */
                                            /* then it is added to the initial shuffling seed and if it is even it is */
                                            /* subtracted. If this is not defined only pure initial seeds are used. */
#define MEMCHECK 1                          /* Defines how to perform the dynamic memory check in Otter(), Discount(), UEQOtter() */
                                            /* and UEQDscnt() to prevent using more memory than the specified limit according to */
                                            /* the following values: */
                                            /* 0 -> Don't perform dynamic memory check. This is the fastest option and can be used */
                                            /*      in CASC competition as StarExec will kill the process if memory is exceeded. */
                                            /* 1 -> Perform dynamic memory check using mallocmn(), reallocmn() and strdupmn functions. */
                                            /*      This is the second faster option. */
                                            /* 2 -> Same as value 1 and also performs dynamic memory check using GetAvailableMemory() */
                                            /*      function. This is safest way to prevent swapping but slows down  the program */
                                            /*      about 1% of elapsed time. */
                                            /* 3 -> Perform dynamic memory check using mallinfo2() function. This is the more accurate */
                                            /*      way to check memory usage. It may be similar than value 2 in terms of time or */
                                            /*      maybe a bit slower. */
//#define TUNINGTIME                          /* If this is defined then pre-processing time will not be accounted for timeouts. */
                                            /* This is used when tuning strategies in StarExec because the timeout is set to */
                                            /* a low value and that is a handicap for problems with large preprocessing time. */
//#define SEMANTICTAUTOLOGY               /* If this is defined then semantic tautology detection is enabled. */

/* Other algorithm parameter definitions. These are candidates to be adjusted via command options. The definitions */
/* of some of them may significantly change the code generation. */
//#define SINGLEANY                       /* If this is defined then the SelectItems() function will select the first */
                                        /* negative literal if selection option is set to SINGLENEG or the first positive */
                                        /* literal if selection is set to SINGLEPOS. If this is NOT defined then the */
                                        /* SelectItems() function will select the MAXIMUM (if it exist) or first MAXIMAL */
                                        /* negative (if SINGLENEG) or positive (if SINGLEPOS) literal with respect to all */
                                        /* the negative literals. The results obtained is that it is better to define SINGLEANY. */
#define DEMODPOSITION 1                 /* If this equals 1 then demodulations are done first in FwdSimplify() function. */
                                        /* If this equals 2 then demodulations are done just after FirstSimplify() call. */
                                        /* If this equals 3 then demodulations are done just after SecondSimplify() call. */
                                        /* Value 1 is a bit better than the other values. */
#define KEEPSATMODEL 1                  /* If this equals 1 then each execution of SatSolver() function keeps the previous model */
                                        /* until a node is found without a forced option match. This is a slightly better option. */
                                        /* If this doesn't equal 1 then each execution of SatSolver() function keeps the previous */
                                        /* model until a node is found without a forced option match and there are still some  */
                                        /* forced options available. */
#define PRECEDENCE 1                    /* If this is zero then the symbol precedence (precedence field in haschain structure) */
                                        /* is set increasingly as new symbols are defined. Otherwise the symbols that appear */
                                        /* more times are assigned a lower precedence. */
//#define TREENODESFIFO                   /* If this is defined then nodes that are not variables will be added FIFO to the peer */
                                        /* nodes chain, otherwise they will be added LIFO. See TREEVARLAST dfinition for the */
                                        /* position of nodes that are variables in the peer nodes chain. */
//#define TREELEAFSFIFO                   /* If this is defined then leafs will be added FIFO to the peer leafs chain, otherwise */
                                        /* they will be added LIFO. */
#define TREEVARLAST                     /* If this is defined then discrimination tree nodes that are variables are always */
                                        /* in the last position of the peer nodes chain. Otherwise they are placed in the */
                                        /* first position. */
//#define UNITCLRESOLUTION                /* If this is defined then a resolution between a unit clause without assertions being */
                                        /* forward simplified will be attempted with active unit clauses without assertions. */
                                        /* It is much better not to use this, but the code has been left just in case it may be */
                                        /* useful in the future. */
#define REDUNDANCYCOMPLETE              /* If this is defined then the "complete" (by now) version of RedundancyChk() function */
                                        /* is used. Otherwise the Vampire inspired original and not complete version is used. */
#define LKAHDWITHUNIFY                  /* If this is defined the lookahead calls to QueryDscTree() are followed by calls to */
                                        /* Unify() to improve the accuracy of the lookahead estimation at the cost of */
                                        /* slowing down the program. */
#define WEIGHTLKAHDCLAUSES 5            /* Integer greater than 1. If problem type is 3 or 8 then for some weight queues the weight */
                                        /* of the first WEIGHTLKAHDCLAUSES is corrected with the lookahead number of the clause. */
#define LRSWEIGHTLIMIT                  /* If this is defined then Limited Resource Strategy is used to limit maximum clause weight. */
                                        /* With layered clause selection this is not as accurate but anyway it can help. */
//#define ENHANCEDDESTEQRES               /* If this is defined then the destructive equality resolution simplification is enhanced */
                                        /* in DestrEqResolution() function by generating a null clause if the input clause is a */
                                        /* unit clause inequality and both root terms unify. */
#define MAXGRJVARS 5                    /* Maximum number of variables in a clause to activate ground joinability for it. If this */
                                        /* is not defined then no limit will be imposed. */

/* Definitions for parse functions return types. */
#define NEGATION 1                          /* Negation. */
#define UNIVERSALQ 2                        /* Valid universal quantifier. */
#define EXISTENTIALQ 3                      /* Valid existential quantifier. */
#define BRACKET 4                           /* Open parenthesis "(". */
#define VALIDATOMIC 5                       /* Valid atomic sentence. */
#define UNDECLAREDTERM 6                    /* Predicate or function not previously declared. */
#define DECLAREDFUNC 7                      /* Term has been previously declared as a function. */
#define DECLAREDPRED 8                      /* Term has been previously declared as a predicate. */
#define STDVARNAME 9                        /* Standardized variable reserved name. */
#define DECLAREDVAR 10                      /* Declared variable. */
#define VALIDVARLIST 11                     /* Valid variable list. */
#define VALIDSENTENCE 12                    /* Valid sentence. */
#define VALIDNAME 13                        /* Valid name. */

/* Define's for syntax error codes. */
#define NOMEMORY 100000100                  /* No memory. */
#define EMPTYSENTENCE 100000101             /* Empty sentence. */
#define EMPTYBLOCK 100000102                /* Empty parenthesis block. */
#define INVTPTPVARNAME 100000103            /* Invalid TPTP variable name. */
#define INVTPTPNAME 100000104               /* Invalid TPTP predicate or function name. */
#define INVALIDATOMIC 100000105             /* Invalid atomic sentence. */
#define INVALIDVARLIST 100000106            /* Invalid variable list. */
#define INVALIDNAME 100000107               /* Invalid or too long variable, constant, predicate or function name. */
#define INVALIDTERM 100000109               /* Invalid term. */
#define MISSINGARGS 100000110               /* Missing arguments. */
#define ALREADYDECLARED 100000111           /* Variable name already declared in equal scope quantifier. */
#define MISSINGVARINLIST 100000112          /* Missing variable in a variable list. */
#define INVALIDDECLAREDARGS 100000113       /* Predicate or function doesn't match previous declared number of arguments. */
#define INVALIDDECLAREDTYPE 100000114       /* Predicate or function doesn't match previous declared type. */
#define TOOMANYARGS 100000115               /* Too many arguments. */
#define RESERVEDNAME 100000116              /* Use of standard variable name for a function or predicate. */
#define SYNTAXERROR 100000117               /* Generic syntax error. */

/* Definitions for flags field of kbase structure. */
#define RETENTIONUSED 1                     /* LRS or weight limit retention processes have been used and some clauses have */
                                            /* been discarded. The KB now is not complete. */

/* Definitions for flags field of cmprefix structure and type argument of simplify functions. */
/* These flags must be compatible with SELECTED flag and also with HORN flag except FROMNOTAXIOM */
/* that may have the same value as HORN flag. The HORN flag is set in clauses if appropriate but */
/* it is not used by the moment. The flags FROMCONJECTURE, FROMAXIOM, FROMNOTAXIOM and FROMMAINFILE */
/* must be compatible with flags in type field of hashchain structure. */
#define TEXTFORMULA 1                       /* Formula is of text type. */
                                            /* These formulas have their own linked queue. */
#define UNPROC 2                            /* Formula is an unprocessed binary clause. */
                                            /* These formulas have their own linked queue. */
#define PASSIVE 4                           /* Formula is a passive binary clause. */
                                            /* These formulas have their own linked queue. */
#define ACTIVE 8                            /* Formula is an active binary clause. */
                                            /* These formulas have their own linked queue. */
#define LOCKED 16                           /* Formula is a locked binary clause. */
                                            /* These formulas have their own linked queue. */
#define NOFACTORS 32                        /* Clause have no factors. */
#define INDISPOSALQUEUE 128                 /* Clause is in disposal queue for later deletion or locking. */
#define FROMPASSIVE 256                     /* Clause is coming from PASSIVE, so only subsumption must be checked */
                                            /* in FirstSimplify() function. This is only used as a flag argument in */
                                            /* some simplifying functions. This is not used in the flags field of the */
                                            /* cmprefix structure. */
#define NUMBERED 512                        /* Clause has been assigned a number. */
#define INPROOFQUEUE 1024                   /* Clause or hashchcmp component is already in proof queue. This flag must be */
                                            /* compatible with hashchcmp structure flags. */
#define ASSERTSCHAINED 2048                 /* Assertions for this clause has been already chained. Only used for */
                                            /* binary clauses. */
#define FROMCONJECTURE 4096                 /* Formula was obtained from a negated conjecture in the pre-processing. */
                                            /* Used in hashchain type field flags for symbols appearing in these formulas. */
#define KEPTCLAUSE 16384                    /* Clause was kept in the relevance filtering pre-processing. */
#define FROMNOTAXIOM 32768                  /* Symbol appears in axiom like formulas hypothesis, definition, assumption, */
                                            /* lemma, theorem and corollary, but not in formulas with axiom role. */
											/* Used only in hashchain type field flags for symbols appearing in these clauses. */
#define FROMMAINFILE 65536                  /* Clause is coming from the main file and can be considered as a conjecture */
                                            /* to the effects of detecting contradictory axioms. */
#define PICKED 131072                       /* Clause has been selected as ACTIVE. Due to AVATAR architecture an active clause */
                                            /* may later be an unprocessed clause so any type of clause may have this flag. */
#define WEIGHTED 262144                     /* Weight, avatar and horn distances have been already assigned to a clause. */
#define FROMAXIOM 524288                    /* Symbol appears in axiom formulas. Used only in hashchain type field flags */
                                            /* for symbols appearing in these clauses. */
#define GROUNDJOINABLE 1048576              /* Formula is a ground joinable binary clause. These formulas have their own */
                                            /* linked queue but use the same top and bottom pointers as UNPROC clauses, that is, */
                                            /* frstunproc and lstunproc fields in kbase structure. */
#define WEAKLYORIENTED 2097152              /* Clause is a binary positive unit equality weakly oriented. */

/* Definitions for inference field of cmprefix structure. */
#define AXIOM 0                             /* Formula is an axiom. */
#define HYPOTHESIS 1                        /* Formula is an hypothesis (axiom like). */
#define DEFINITION 2                        /* Formula is a definition (axiom like). */
#define ASSUMPTION 3                        /* Formula is an assumption (axiom like). */
#define NEGCONJ 4                           /* Formula is a negated conjecture. This is also used as a flag and must be */
                                            /* compatible with FROMMAINFILE above. */
#define CONJECTURE 5                        /* Formula was an input conjecture to the problem. */
#define LEMMA 6                             /* Formula is a lemma (axiom like, redundant). */
#define THEOREM 7                           /* Formula is a theorem (axiom like, redundant). */
#define COROLLARY 8                         /* Formula is a corollary (axiom like, redundant). */
#define PLAIN 9                             /* Formula has a plain role, accepted as an axiom just in case . */
#define CLAUSIFY 10                         /* Formula comes from clausifying an input formula. */
#define RESOLUTION 11                       /* Formula was inferred by resolution. */
#define PARAMODULATION 12                   /* Formula was inferred by paramodulation. */
#define FACTORING 13                        /* Formula was inferred by factoring. */
#define FWDEMODULATION 14                   /* Formula was inferred by forward demodulation. */
#define BKDEMODULATION 15                   /* Formula was inferred by backward demodulation. */
#define REMOVEDUPS 16                       /* Formula is the result of removing duplicate literals from another formula. */
#define TRIVEQRES 17                        /* Formula was inferred by trivial equality resolution. */
#define EQRESOLUTION 18                     /* Formula was inferred by equality resolution. */
#define DESTEQRES 19                        /* Formula was inferred by destructive equality resolution. */
#define EQFACTORING 20                      /* Formula was inferred by equality factoring. */
#define FWSUBSRESLTNS 21                    /* Formula was inferred by forward subsumption resolution. */
#define BKSUBSRESLTNS 22                    /* Formula was inferred by backward subsumption resolution. */
#define SPLIT 23                            /* Formula was inferred by clause splitting. */
#define COMPONENT 24                        /* Formula is a clause component. */
#define CONTRADICTION 25                    /* Formula was inferred by first order solver contradiction. */
#define PRENNFCONV 26                       /* Formula is the result of a pre-nnf conversion. */
#define FORMULARENAMING 27                  /* Formula is the result of renaming a subformula. */
#define PREDICATEDEF 28                     /* Formula is a predicate definition for subformula renaming. */
#define NNFCONV 29                          /* Formula is the result of a full-nnf conversion. */
#define MINISCOPING 30                      /* Formula is the result of miniscoping. */
#define SKOLEMIZATION 31                    /* Formula is the result of skolemization. */
#define TFSIMPLIFICATION 32                 /* Formula is the result of a $true and $false simplification. */
#define EQUALITYSPLIT 33                    /* Formula is the result of a positive unit equality split. */

/* Definitions for byte markers in compiled clauses and discrimination trees, type field of hashchain */
/* structures and type parameter of CreateGlobalSymbol() function. These flags must be compatible */
/* with FROMCONJECTURE, FROMAXIOM, FROMNOTAXIOM and FROMMAINFILE also used in type field of hashchain */
/* structures for learning process. PREDICATE and FUNCION are also used in flags field of nameschain structure. */
#define UNITEND 1                       /* End of clause marker. Only used for byte markers. */
                                        /* This flag must be compatible with flags for Generalize() and */
                                        /* Unify() functions and with flags in hashchcmp and satnode structures */
                                        /* and assertions signbyte (see below). */
#define EQUALITY 2                      /* Item is an equality literal. Also used in UpdateSymbolParms() function flags. */
#define PREDICATE 4                     /* Item is a predicate literal. Also used in nameschain structure flags. */
#define NEGATED 8                       /* Item is a negated literal. Only used for byte markers and in */
                                        /* UpdateSymbolParms() function flags. */
                                        /* This flag must be compatible with ORIENTEDEQ flag and flags for Generalize() */
                                        /* and Unify() functions (see below). */
#define VARIABLE 16                     /* Item is a variable. Only used for byte markers and nameschain structure flags. */
#define FUNCTION 32                     /* Item is a function (may be a constant, i.e. arity zero). */
#define SELECTED 64                     /* Item is a selected literal or equality root term. */
                                        /* This flag must be compatible with flags for QueryGeneric() function. */
#define MAXIMUM 128                     /* Item is a maximum equality root term. This must be different than ITEM1SEC, STANDARD */
                                        /* and NONRECURSIVE flags. */
#define PMSELECTED 128                  /* If this is set to the same value as MAXIMUM then a positive equality will be selected */
                                        /* for paramodulating from it only if it is in a clause without negative literals. */
                                        /* If this is set to the same value as SELECTED then the above condition will not */
                                        /* be required. It works a bit better setting it to the same value as MAXIMUM. */
//#define DEFINEDPRED 256                 /* Used only in "type" parameter of CreateGlobalSymbol() function */
                                        /* and type field of hashchain structure. If this is commented out then predicate */
                                        /* definitions coming from sub-formula renaming will not have a special treatment */
                                        /* for literal ordering purposes. It seems better to comment it out. */
#define RELEVANT 1024                   /* Symbol is in a relevant clause. Used only in type field of hashchain */
                                        /* structure for clause relevance filtering. */
#define SKOLEM 2048                     /* Symbol is a skolem. Used only in type field of hashchain structure for */
                                        /* learning  process. */
#define DISTINCTOBJECT 8192             /* Symbol is a "distinct object". Used only in type field of hashchain structure */
                                        /* and type parameter of CreateGlobalSymbol() function. */

/* Definitions for "type" parameter in NextItem() function. */
#define OVERSUBTERMS 1                  /* Jump to next item that is not a subterm of the given item. */
#define IMMED 2                         /* Jump to item immediately following the given item. */

/* Definitions for Unify() and Generalize() flags. */
/* These flags must be compatible with the UNITEND and NEGATED flags (see above). */
#define LIST 2                          /* Parts to be unified are a list of terms. */
#define ITEM1SEC 16                     /* The first Unify() function argument or the variable in a substitution */
                                        /* belongs to a secondary clause. They belong to the main clause by default. */
                                        /* Also used for flag value of CheckEqRootOrder() function so it must be */
                                        /* different than MAXIMUM, STANDARD and NONRECURSIVE values defined elsewhere. */
#define ITEM2SEC ITEM1SEC<<1            /* The second Unify() function argument or the substituting term in a */
                                        /* substitution  belongs to a secondary clause. They belong to the main */
                                        /* clause by default. This must be ITEM1SEC shifted by 1 bit. */
#define INITSUBST 64                    /* Initialize substitution to null. */
#define DEBUG 128                       /* Call BuildTxtSubstData() function to show results in text form. */

/* Definitions for status fields of kbase and mpctl structures. They must be compatible with NOMEMORY above. */
#define STOPPED 0                       /* Process is stopped. */
#define RUNNING 1                       /* Process is running a (possibly multi-processing) first order solver algorithm */
                                        /* with or without a SAT solver. */
#define RUNNINGSAT 2                    /* Process is running a single process SAT solver algorithm. */
#define SATISFIABLE 3                   /* Formulas in KB are satisfiable, if there are conjectures */
                                        /* they are not refutable. */
#define UNSATISFIABLE 4                 /* Formulas in KB are unsatisfiable, if there are conjectures */
                                        /* they are refutable. */
#define UNKNOWN 5                       /* It is unknown if formulas in KB are satisfiable or unsatisfiable. */
#define TIMEOUT 6                       /* Saturation algorithm timed out. This cannot be 0 or 1. */
#define SATQTIMEOUT 7                   /* A timeout was produced in SatQ2SatSolver() function. */

/* Definitions for ordtype values and CompareTerms() flag. */
/* IMPORTANT: If additional ordering types are added then MAXLEARNRECORDS definition must be changed. */
#define STANDARD 1                      /* Use standard KBO ordering. */
#define NONRECURSIVE 2                  /* Use non recursive KBO ordering. This flag must be compatible  with */
                                        /* MAXIMUM and ITEM1SEC flags defined elsewhere. */
#define LEXICOGRAPHIC 4                 /* Use >> symbol precedence relation for atom ordering. */

/* Definitions for demodulation and destructive equality resolution options (DEMODULAITON option). */
#define DEMODON 0                       /* Demodulation is enabled without restrictions. */
#define DEMODOFF 1                      /* Demodulation is disabled. */
#define DEMODCPL 2                      /* Demodulation and destructive equality resolution are restricted */
                                        /* to cases that don't compromise completeness. */

/* Definitions for weight functions. */
/* IMPORTANT: If additional types of weight dunctions are added then MAXLEARNRECORDS definition must be changed. */
#define UNIFORM 0                       /* Use a uniform weight function assigned a weight=1 to all functions. */
#define ARITY 1                         /* Use weight function = 1 + 10·arity(function). */

/* Definitions for ordering functions return values and for the result field of the */
/* equpnode structure. See the functions description for the meaning of each one, */
/* as it may differ from one function to another. */
#define EQUAL 1                         /* item1 = item2. */
#define GREATER 2                       /* item1 > item2. */
#define LESSER 4                        /* item2 > item1. */
#define NOGREATER 8                     /* item2 ·> item1 but neither item2 = item1 nor item2 > item1. */
#define NOLESSER 16                     /* item1 ·> item2 but neither item1 = item2 nor item1 > item2. */
#define FAILURE 32                      /* None of the above. */
#define EQUALW 64                       /* weight(item1) = weight(item2). */
#define FAILUREW 128                    /* weight(item1) is not comparable with weight(item2). */

/* Definitions for algorithm field of kbase structure. */
/* IMPORTANT: If additional algorithms are added then MAXLEARNRECORDS definition must be changed. */
#define OTTER 2                         /* Otter algorithm. */
#define DISCOUNT 3                      /* Discount algorithm. */
#define UEQDSC 4                        /* Unfailing completion Discount algorithm. */
#define UEQOTT 5                        /* Unfailing completion Otter algorithm. */

/* Definitions for EditDiscrTrees() function flag parameter. */
#define ADDTREES 1                      /* Add tree branches for specified clause. */
#define DELETETREES 2                   /* Delete tree branches for specified clause. */

/* Definitions for type argument of QueriDscTree() function. */
#define UNIFICATION 0                   /* Search is for an item that unifies with the query term. */
#define INSTANCE 1                      /* Search is for an item that is an instance of the query term. */
#define GENERALIZATION 2                /* Search is for an item that is a generalization of the query term. */

/* Definitions for flags of AddBinClause2KB() function. */
#define SELECT 1                        /* Appropriate clause items will be selected. */
#define SETFNUMBER 2                    /* Set formula number according to the nformulas field of KB. */

/* Definitions for flags of JumpOverBlockOrTerm() function. */
#define TOEND 0                         /* Jump to the last item in the block or term. */
#define PASTEND 1                       /* Jump to the item following the last item in the block or term. */

/* Definitions for mxwghtopt field of options structure. */
#define MXWEIGHTOFF 1                   /* Maximum weight option deactivated for algorithms without LRS enabled. */
#define MXWEIGHTLRS 2                   /* Maximum weight calculated using Limited Resource Strategy enabled */
                                        /* for all algorithms. */
#define MXWEIGHTLRSOTT 4                /* Maximum weight calculated using Limited Resource Strategy enabled for Otter */
                                        /* algorithm, disabled for Discount algorithm. */
#define MXWEIGHTLRSDSC 8                /* Maximum weight calculated using Limited Resource Strategy enabled for Discount */
                                        /* algorithm, disabled for Otter algorithm. */
#define MXWEIGHTLRSOFF 16               /* Maximum weight calculated using Limited Resource Strategy id disabled */
                                        /* for all algorithms. */
#define MXWEIGHTNUM 32                  /* Maximum weight selected by user for algorithms without LRS enabled. */

/* Definitions for SYNTAX and PROOF options. */
#define TPTP 1                          /* TPTP mode syntax. */
#define DRODI 2                         /* Drodi mode syntax. */
#define PROOFOFF 3                      /* Don't print proof. */

/* Definitions for program modes (pgmmode global variable). */
#define NONINTERACTIVE 0                /* non interactive mode, entered if some arguments are given to the program. */
#define INTERACTIVE 1                   /* Interactive mode, entered if no arguments are given to the program. */

/* Definitions for select field of options structure. The definitions are set as if they were flags */
/* to speed up some tests, but only one of the values (flags) is allowed. */
/* IMPORTANT: If additional SELECT types are added then MAXLEARNRECORDS definition must be changed. */
#define MAXIMAL 1                       /* All maximal literals are selected. */
#define SINGLENEG 2                     /* If there are negative literals then the first one is selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define MULTINEG 4                      /* If there are negative literals then the negative literals */
                                        /* that are maximal in the set of negative literals are selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define SINGLEPOS 8                     /* If there are positive literals then the first one is selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define MULTIPOS 16                     /* If there are positive literals then the positive literals */
                                        /* that are maximal in the set of positive literals are selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define SELECTALL 32                    /* Select all literals. */
#define MXSINGLENEG 64                  /* Equivalent to strategy 1 of document Selecting the Selection by Hoder. */
#define TYPE1 128                       /* Equivalent to strategy 1002 of document Selecting the Selection by Hoder. */
#define TYPE2 256                       /* Equivalent to strategy 1003 of document Selecting the Selection by Hoder. */
#define TYPE3 512                       /* Equivalent to strategy 1004 of document Selecting the Selection by Hoder. */
#define TYPE4 1024                      /* Equivalent to strategy 1010 of document Selecting the Selection by Hoder. */
#define TYPE5 2048                      /* Equivalent to strategy 1011 of document Selecting the Selection by Hoder. */
#define TYPE6 4096                      /* Equivalent to strategy 1012 of document Selecting the Selection by Hoder. */
#define TYPE1CPL 8192                   /* Complete version of TYPE1. */
#define TYPE2CPL 16384                  /* Complete version of TYPE2. */
#define TYPE3CPL 32768                  /* Complete version of TYPE3. */
#define TYPE4CPL 65536                  /* Complete version of TYPE4. */
#define TYPE5CPL 131072                 /* Complete version of TYPE5. */
#define TYPE6CPL 262144                 /* Complete version of TYPE6. */

/* define for complete selections mask. */
#define COMPLETESELECTIONS (MAXIMAL|SINGLENEG|MULTINEG|TYPE1CPL|TYPE2CPL|TYPE3CPL|TYPE4CPL|TYPE5CPL|TYPE6CPL)

/* Definitions for flags field in hashchcmp and satnode structures and assertions signbyte */
/* (see AddAssertions() function comments). These definitions must be compatible with UNITEND */
/* that is also used in assertions signbyte and with INPROOFQUEUE cmprefix structure flag */
/* that is also used in hashchcmp flags. */
#define POSITIVE 2                      /* The component is true in SAT interpretation model or in assertion_data. */
#define NEGATIVE 4                      /* The component is false in SAT interpretation model or in assertion_data. */
#define UNDEFINED 8                     /* The component is undefined in SAT interpretation model. This is only used */
                                        /* in flags field of hashchcmp structure. */
#define GROUND 16                       /* The component is ground. This is only used in flags field of hashchcmp structure. */
                                        /* This flag must be compatible with flags for pbmflags global variable and serves */
                                        /* to identify ground problems. */
#define PRPOSITIVE 32                   /* The component was true in previous model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define PRNEGATIVE 64                   /* The component was false in previous model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define PRUNDEFINED 128                 /* The component was undefined in previous model of SAT interpretation. This is only used */
                                        /* in flags field of hashchcmp structure. */
#define LASTOPTION 256                  /* The assignment of node is the last possibility available, no other possibility */
                                        /* (true or false) is left or valid to be explored.*/
#define ORDERED 512                     /* The hashchcmp or satclause structure element has been linked to the ordered */
                                        /* ordered sub-queue. */
#define FREENODE 2048                   /* The node is allocated but it is not in use and its data is no longer necessary. */
                                        /* The node can be reused. This way unnecessary memory allocation and freeing is */
                                        /* avoided. This is only used in flags field of satnode structure. */
#define MODELPOSITIVE 4096              /* The component is true in current model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define MODELNEGATIVE 8192              /* The component is false in current model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define MODELUNDEFINED 16384            /* The component is undefined in current model of SAT interpretation. This is only used */
                                        /* in flags field of hashchcmp structure. */
#define POSCOMPADDED 32768              /* The component clause C <- [C] has been added to first order solver. This is */
                                        /* only used in flags field of hashchcmp structure. */
#define NEGCOMPADDED 65536              /* The component clause ~C <- ~[C] has been added to first order solver. This is */
                                        /* only used in flags field of hashchcmp structure. */

/* Definitions for flags in pbmflags global variable. The GROUND flag defined above is also used for pbmflags. */
#define EXISTCONJ 1                     /* A conjecture exist for the current problem. */
#define EXISTNEGCONJ 2                  /* A negated conjecture exist for the current problem. */
#define WITHHORNCLAUSES 4               /* Problem is not horn but there are horn clauses. */
#define UNITEQUALITY 8                  /* Problem has some positive unit equalities. */
#define NONUNITEQUALITY 32              /* Problem has some equalities none of which are positive unit equalities. */
#define UNITCLAUSE 64                   /* All clauses in problem are unit clauses. */
#define NOTCAX 128                      /* The proof includes some formulas from main file and therefore has not */
                                        /* contradictory axioms. */
#define SATREFUTATION 256               /* There was a Sat refutation. */
#define HORN 32768                      /* Problem is horn. This flag must be compatible with flags field */
                                        /* of cmprefix structure. */
#define CNFFORMAT 65536                 /* Problem was read in CNF TPTP format. */
#define EXISTAXIOMFILE 131072           /* Problem included an axiom file. */
#define EXISTAXIOMLIKE 262144           /* Problem has axiom or axiom like formulas. */
#define TYPE8ELIGIBLE 524288            /* All clauses are unit equalities except a negated conjecture that is an inequality. */

/* Definitions for type field in proofnode structure. */
#define SYMBOLDEFNTN 1                  /* Symbol definition. */
#define SATCLAUSE 2                     /* Sat clause. */
#define ACLAUSE 3                       /* A-Clause. */

/* Definitions for flags field in lkedfitem structure. */
#define OPENBRACKET 1                   /* Opening bracket type item. */
#define CLOSEBRACKET 2                  /* Closing bracket type item. */
#define EQUPRED 4                       /* Equality predicate type item. */
#define NAMEDPRED 8                     /* Normal predicate type item. */
#define NAMEDFUNC 16                    /* Function type item. */
#define NAMEDVAR 32                     /* Variable type item. */
#define UQUANTIFIER 64                  /* Universal quantifier type item. */
#define EQUANTIFIER 128                 /* Existential quantifier type item. */
#define ANDOPER 256                     /* AND connective type item. */
#define OROPER 512                      /* OR connective type item. */
#define IMPOPER 1024                    /* Implication connective type item. */
#define REVIMPOPER 2048                 /* Reverse implication connective type item. */
#define EQUIVOPER 4096                  /* Equivalence connective type item. */
#define XOROPER 8192                    /* XOR connective type item. */
#define NOROPER 16384                   /* NOR connective type item. */
#define NANDOPER 32768                  /* NAND connective type item. */
#define NEGITEM 65536                   /* Negated item. */
#define POSPOL 131072                   /* Positive polarity item. */
#define NEGPOL 262144                   /* Negative polarity item. */
#define NEUTRALPOL 524288               /* Neutral polarity item. */
#define RENAMED 1048576                 /* Variable renamed in a quantifier or variable item. */
#define FREEVAR 2097152                 /* The quantifier variable is a free variable in the renamed block. */
#define TRUEPRED 4194304                /* The predicate is a $true predefined predicate. */
#define FALSEPRED 8388608               /* The predicate is a $false predefined predicate. */

/* Definitions for flags parameter of IsConnected() function. */
#define SUBSTITUTION2 1                 /* Use a₁≻a₂≻ ... ≻aₚ ground substitution for variables. */
#define MXTERMCHECKED 2                 /* mxterm argument has been checked for being LESSER than the */
                                        /* original critical pair top term. */
#define MNTERMCHECKED 4                 /* mnterm argument has been checked for being LESSER than the */
                                        /* original critical pair top term. */

/* Definitions for childtype field of dstrnode structure. */
#define NODECHILD 0                     /* Child is a tree node (dstrnode structure). */
#define LEAFCHILD 1                     /* Child is a tree leaf (dstrleaf structure). */

/* Definitions for clause relevance filtering program option (prproflags global variable). */
#define CLFILTERINGRESIGN 1             /* Clause relevance filtering is not performed performed in case of timeout condition. */
#define CLFILTERINGOFF 2                /* No clause relevance filtering is performed. */
#define CLFILTERINGON 4                 /* Clause relevance filtering is always performed. */
#define CLFILTERINGNONEQU 8             /* Clause relevance filtering is performed in problems without equalities only. */
                                        /* Only one of the CLFILTERINGOFF, CLFILTERINGON or CLFILTERINGNONEQU flags can be set. */
                                        /* However CLFILTERINGRESIGN flag is compatible with the other three flags. */
#define CLFILTERINGITERLIMIT 16         /* The threshold in ppthrshld global variable is the maximum number of iterations */
                                        /* allowed in the relevance filtering algorithm. Otherwise the threshold in ppthrshld */
                                        /* global variable is the maximum time in seconds allowed for the algorithm. */
#define CLFILTERINGLIMITREACHED 32      /* A timeout condition occurred or maximum number of iterations reached during */
                                        /* relevance filtering pre-processing before */
                                        /* the filtering process was completed. */
#define CLFILTERINGPRUNNED 64           /* The relevance filtering deleted one or more clauses. */
#define CLFILTERINGTIMEOUT 0.1          /* Default timeout for relevance filtering pre-processing. */

/* Definitions of masks for each type of adjustable parameter. */
#define ALGMASK 0b100000000000000000001         /* ALGORITHM parameter mask. */
#define SELECTMASK 0b11001000000000110          /* SELECT parameter mask. */
#define LAYERMASK 0b1000                        /* Weight factor index parameter mask. */
#define DEMODMASK 0b100000000010000             /* Demodulation parameter mask. */
#define FACTORINGMASK 0b100000                  /* FACTORING parameter mask. */
#define TERMORDRMASK 0b1000000                  /* TERMORDER parameter mask. */
#define LITORDRMASK 0b110000000                 /* LITORDER parameter mask. */
#define FUNCWMASK 0b1000000000                  /* FUNCTIONW parameter mask. */
#define MAXWMASK 0b110000000000                 /* MAXWEIGHT parameter mask. */
#define LOOKAHEADMASK 0b100000000000000000      /* LOOKAHEAD parameter mask. */
#define RATIO2MASK 0b1000000000000000000        /* SELECTRATIO2 parameter mask. */
#define GAMMAMASK 0b10000000000000000000        /* GAMMA parameter mask. */
#define GRJOINMASK 0b11000000000000000000000    /* GRJOIN parameter mask. */
#define CONNECTMASK 0b100000000000000000000000  /* CONNECT parameter mask. */
#define WEAKRWMASK 0b1000000000000000000000000  /* WEAKRW parameter mask. */

/* Definitions of values for each type of adjustable parameter. */
#define OTTER_P 0                                /* Otter algorithm. */
#define DISCOUNT_P 1                             /* Discount algorithm. */
#define UEQDSC_P 0b100000000000000000000         /* Unfailing completion Discount algorithm. */
#define UEQOTT_P 0b100000000000000000001         /* Unfailing completion Otter algorithm. */
#define SINGLENEG_P 0                            /* SINGLENEG SELECT option. */
#define MULTINEG_P 0b10                          /* MULTINEG SELECT option. */
#define MAXIMAL_P 0b100                          /* MAXIMAL SELECT option. */
#define SINGLEPOS_P 0b110                        /* SINGLEPOS SELECT option. */
#define MULTIPOS_P 0b1000000000000               /* MULTIPOS SELECT option. */
#define LAYER1_P 0                               /* Layer 1 option. */
#define LAYER2_P 0b1000                          /* Layer 2 option. */
#define DEMODON_P 0                              /* Demodulation is ON. */
#define DEMODOFF_P 0b10000                       /* Demodulation is OFF. */
#define FACTON_P 0                               /* ON FACTORING option. */
#define FACTOFF_P 0b100000                       /* OFF FACTORING option. */
#define TRMORDSTD_P 0                            /* STANDARD TERMORDER option. */
#define TRMORDNR_P 0b1000000                     /* NONRECURSIVE TERMORDER option. */
#define LITORDSTD_P 0                            /* STANDARD LITORDER option. */
#define LITORDNR_P 0b10000000                    /* NONRECURSIVE LITORDER option. */
#define LITORDLEX_P 0b100000000                  /* LEXICOGRAPHIC LITORDER option. */
#define UNIFORM_P 0                              /* UNIFORM FUNCTIONW option. */
#define ARITY_P 0b1000000000                     /* ARITY FUNCTIONW option. */
#define LRSOTT_P 0                               /* LRSOTTER MAXWEIGHT option. */
#define LRSOFF_P 0b10000000000                   /* LRSOFF MAXWEIGHT option. */
#define LRSALL_P 0b100000000000                  /* LRS MAXWEIGHT option. */
#define LRSDSC_P 0b110000000000                  /* LRSDISCOUNT MAXWEIGHT option. */
#define LEARN_P 0b10000000000000                 /* Use learning option. This doen't need an adjustable parameter mask. */
#define DEMODCPL_P 0b100000000000000             /* Redundancy check is on for demodulation and destructive equality resolution. */
#define SELECTALL_P 0b1000000000010              /* SELECTALL SELECT option. */
#define MXSINGLENEG_P 0b1000000000100            /* MXSINGLENEG SELECT option. */
#define TYPE1_P 0b1000000000110                  /* TYPE1 SELECT option. */
#define TYPE2_P 0b1000000000000000               /* TYPE2 SELECT option. */
#define TYPE3_P 0b1000000000000010               /* TYPE3 SELECT option. */
#define TYPE4_P 0b1000000000000100               /* TYPE4 SELECT option. */
#define TYPE5_P 0b1000000000000110               /* TYPE5 SELECT option. */
#define TYPE6_P 0b1001000000000000               /* TYPE6 SELECT option. */
#define TYPE1CPL_P 0b1001000000000010            /* TYPE1CPL SELECT option. */
#define TYPE2CPL_P 0b1001000000000100            /* TYPE2CPL SELECT option. */
#define TYPE3CPL_P 0b1001000000000110            /* TYPE3CPL SELECT option. */
#define TYPE4CPL_P 0b10000000000000000           /* TYPE4CPL SELECT option. */
#define TYPE5CPL_P 0b10000000000000010           /* TYPE5CPL SELECT option. */
#define TYPE6CPL_P 0b10000000000000100           /* TYPE6CPL SELECT option. */
#define LKAHEADOFF_P 0                           /* Lookahead is off. */
#define LKAHEADON_P 0b100000000000000000         /* Lookahead is on. */
#define GRJOINOFF_P 0                            /* Ground joinability is off. */
#define GRJOINON_P 0b1000000000000000000000      /* Ground joinability is on. */
#define GRJOINMED_P 0b11000000000000000000000    /* Ground joinability is set to med. */
#define CONNECTOFF_P 0                           /* Connectedness is off. */
#define CONNECTON_P 0b100000000000000000000000   /* Connectedness is on. */
#define WEAKRWOFF_P 0                            /* Weak rewrite is off. */
#define WEAKRWON_P 0b1000000000000000000000000   /* Weak rewrite is on. */

/* Definitions to be used if VERIFYSATSOLVER is not zero. They are always defined anyway. */
#define SATOK 0                         /* SAT solver status is OK. */
#define SATSAT 1                        /* SAT solver has reported a refutation but all SAT clauses are satisfied. */
#define SATUNSAT 2                      /* SAT solver has reported a model but not all SAT clauses are satisfied. */
#define SATUNPR1 4                      /* There are unprocessed clauses that must be locked. */
#define SATUNPR2 8                      /* There are clauses in unprocessed queue not flagged as unprocessed. */
#define SATUNPR12 16                    /* Conditions SATUNPR1 and SATUNPR2. */
#define SATPASSIVE1 32                  /* There are passive clauses that must be locked. */
#define SATPASSIVE2 64                  /* There are clauses in passive queue not flagged as passive. */
#define SATPASSIVE12 128                /* Conditions SATPASSIVE1 and SATPASSIVE2. */
#define SATACTIVE1 256                  /* There are active clauses that must be locked. */
#define SATACTIVE2 512                  /* There are clauses in active queue not flagged as active. */
#define SATACTIVE12 1024                /* Conditions SATACTIVE1 and SATACTIVE2. */
#define SATLOCKED1 2048                 /* There are active clauses that must be locked. */
#define SATLOCKED2 4096                 /* There are clauses in active queue not flagged as active. */
#define SATLOCKED12 8192                /* Conditions SATLOCKED1 and SATLOCKED2. */
#define SATUNK 16384                    /* SAT solver status is unknown because the problem was not reported as solved. */

/* Definitions related to clause selections. */
#define VARWEIGHT 10                    /* Relative weight of variable symbols for clause weight evaluation. */
#define SYMBOLWEIGHT 20                 /* Relative weight of predicate and function symbols for clause weight evaluation. */
#define SINESEGMENT1 0                  /* End of first segment for SInE selection layer. */
#define SINESEGMENT2 1                  /* End of second segment for SInE selection layer. */
#define AVSEGMENT1 0                    /* End of first segment for Avatar selection layer. */
#define HRNSEGMENT1 0                   /* End of first segment for horn selection layer. */
#define SIAVLYR 0                       /* Specialized layers are SInE and Avatar. This is a possible value of layer */
                                        /* and layerset[0] and layerset[1] fields of options structure. */
#define SIHRNLYR 1                      /* Specialized layers are SInE and Horn. This is a possible value of layer */
                                        /* and layerset[0] and layerset[1] fields of options structure. */
#define AVHRNLYR 2                      /* Specialized layers are Avatar and Horn. This is a possible value of layer, */
                                        /* layerset[0] and layerset[1] fields of options structure. */
#define SINELYR 3                       /* Specialized layer is SInE. This is a possible value of layer, */
                                        /* layerset[0] and layerset[1] fields of options structure. */
#define AVLYR 4                         /* Specialized layer is Avatar. This is a possible value of layer, */
                                        /* layerset[0] and layerset[1] fields of options structure. */
#define HRNLYR 5                        /* Specialized layer is Horn. This is a possible value of layer, */
                                        /* layerset[0] and layerset[1] fields of options structure. */
#define NONELYR 6                       /* No specialized layer. This is a possible value of layer, */
                                        /* layerset[0] and layerset[1] fields of options structure. */

/* Definitions for the automatic learning process. */
#define VECTORLENGTH 78888              /* Number of elements in a clause vector used for learning. */
#define PENALTYPARAMETER 0.1            /* Penalty parameter for learning process. */
#define TOLERANCE1 2.0                  /* Duality gap tolerance used while good and bad clauses accuracy */
                                        /* is unbalanced in dual DCDescentMethod() function. */
#define TOLERANCE2 0.5                  /* Duality gap tolerance used when good and bad clauses accuracy */
                                        /* is balanced in dual DCDescentMethod() function. */
#define MAXRECSITER 0x7fffffff          /* Maximum value of number of iterations times number of records */
                                        /* in dual DCDescentMethod() function. This value should be set */
                                        /* to a value of 84000000 or greater. */
#define BADCLAUSEWEIGHT 100             /* Weight penalty if the clause is not classified as good by the learning algorithm. */
#define GOODCLAUSEWEIGHT 1              /* Weight penalty if the clause is classified as good by the learning algorithm. */
#define GENERICSYMBOLS 16               /* Number of generic symbols for both predicate and functions. */
#define ACCBALANCE 0.95                 /* Good/bad clauses accuracy balance threshold to accept a SVM result without */
                                        /* increasing the number of good clauses incorrectly classified. */
#define PROCLEARNMINMEM 400000000       /* Minimum memory in bytes to perform PROCLEARN process by LearnFromProblem() function. */
#define PROCLEARNMINMEM1 100000         /* Minimum memory in bytes to attempt duplicate some misclassified good clauses. */
#define LEARNMEMORYCHUNK 100            /* Number of items to be (re)allocated each time to fvidx global memory. */
#define MAXLEARNRECORDS 912             /* Maximum number of records per problem type. This is the maximum value of numlrnrecs */
                                        /* global variable. */
                                        /* IMPORTANT: This must be adjusted if the number of possible strategy parameters */
                                        /* for ALGORITHM, SELECT, TERMORDER, LITORDER or FUNCTIONW options changes. */
#define OFFSET1 77778                   /* Offset of maximum symbol depth feature. */
#define OFFSET2 22032                   /* Offset of horizontal features with arity greater than 3. */
#define OFFSET3 22041                   /* Offset of equality horizontal feature. */
#define OFFSET4 22212                   /* Offset of horizontal features with arity 1. */
#define OFFSET5 22374                   /* Offset of horizontal features with arity 2. */
#define OFFSET6 25290                   /* Offset of horizontal features with arity 3. */
#define OFFSET7 78814                   /* Symbol count offset. */
#define OFFSET8 78884                   /* Clause length and literal count offset. */

/* Miscellanea definitions. */
#define VERSION "Drodi V3.6.0"          /* Program version. */
#define SPREAD 0                        /* Use spread core affinity bind option. */
#define CLOSE 1                         /* Use close core affinity bind option. */
#define LRSTHRESHOLD 200                /* Number of passive clauses moved to active per LRS cycle. */
#define LRSCOEF 1.1                     /* Coefficient to apply to number of LRS reachable clauses. */
#define TIMEOUTCCLDSP 10                /* Number of cycles for timeout check in clause disposal. */
#define INPUT_CHUNK_SIZE 1024           /* Chunk size in bytes for command input. */
#define VAR_CHUNK_SIZE 100              /* Chunk size in bytes for text formula names. */
#define STR_CHUNK_SIZE 200              /* Chunk size in bytes for text buffers. */
                                        /* This must be bigger or equal to 10. */
#define TREEPTR_CHUNK_SIZE 4            /* Pointer set chunk size in tree pointer elements. */
#define SUBST_CHUNK_SIZE 1000           /* Chunk size in bytes for substitution buffer. */
#define CLAUSE_CHUNK_SIZE 200           /* Compiled auxiliary clauses chunk size in bytes. */
#define PNDARGS_CHUNK_SIZE 10           /* Pending arguments elements chunk size in number of elements. */
#define VARRENUM_CHUNK_SIZE 50          /* Variable renumbering buffer size in number of elements. */
#define ABSMAXWEIGHT 0xffffffff         /* Absolute maximum clause weight. This is to prevent exhausting the stack space by */
                                        /* too many calls to QueryGeneric() for too long clauses. It is rare but it may happen. */
#define MAXWEIGHT 5000                  /* Maximum number of weight queue pointers hold in each weight queue. */
#define SELECTQUEUES 6                  /* Number of clause selection queues. */
#define MAXPOSITIONS 400                /* Maximum number of hash values for item positions. */
#define MAXSTRATEGIES 4320              /* Number of predefined strategies for problem types 0, 1, 2, 4 and 5. This depends */
                                        /* on the number of parameters that are automatically adjusted depending on the problem */
                                        /* type. See InitStrategies() function for details. */
#define MAXSTRATEGIES3 576              /* Number of predefined strategies for problem type 3. See MAXSTRATEGIES comments above. */
#define MAXSTRATEGIES67 1440            /* Number of predefined strategies for problem types 6 and 7. See MAXSTRATEGIES comments above. */
#define MAXSTRATEGIES8 1728             /* Number of predefined strategies for problem type 8. See MAXSTRATEGIES comments above. */
#define STRATEGIES8UEQ 1152             /* Number of predefined strategies with ALGORITHM set to UEQOTT or UEQDSC for problem type 8. */
                                        /* This is always less than MAXSTRATEGIES8. See MAXSTRATEGIES comments above. */
#define BKSTRATEGIESPERPROC 8           /* Number of backup strategies per process to cover early returns due to incompleteness. */
#define TUNEDSTRATEGIES 200             /* Number of tuned strategies for each problem type. */
#define PBMTYPES 9                      /* Number of problem types. Maximum value is 255 due to limitations of the header field */
                                        /* in learnrecord structure. */
#ifdef UNITCLRESOLUTION
#define DSCTREETYPES 8                  /* Number of discr[] set field elements in hashchain structure. */
#else
#define DSCTREETYPES 6                  /* Number of discr[] set field elements in hashchain structure. */
#endif
#define HASHVALUES 0x10000              /* Number of different hash values for symbol names and clause variants. */
#define MAXMINISCOPETIME 1.0            /* Miniscope time limit in seconds. */
#define EXITWAITSECONDS 1               /* Number of seconds to wait for exit of a non solution winner children process. */
                                        /* Child exit is indicated by clearing procsem[] process specific semaphore. */
#define KILLWAITITERATIONS 20           /* Number of iterations to wait for state changes in child processes via waitpid(). */
#define KILLWAITNANOSECS 10000000       /* Number of nanoseconds to wait between iterations for state changes in child */
                                        /* processes via waitpid(). This must be less than 1000000000. */
#define KILLWAITNANOSECS2 100000000     /* Number of nanoseconds to wait before waiting for state changes in child processes */
                                        /* via waitpid() if an interruption occurs. This must be less than 1000000000. */
#define TMOWAITMICROSECS 500000         /* Additional time in microseconds to wait for timeouts not caught by normal means. */
                                        /* This must be less than 1000000. */
#define MAXUNPROCSHUFFLING 100          /* Maximum number of leading or trailing unprocessed clauses to shuffle. */
#define MEMORYLIMIT 5000000             /* Minimum current available memory in bytes for Saturate() to start processing. */
#define MEMORYLIMIT2 16275932000        /* Parameter used to estimate available memory. See GetAvailableMemory() function */
                                        /* for additional information. */
#define MEMORYMARGIN 1700000000         /* Parameter used to estimate available memory. See GetAvailableMemory() function */
                                        /* for additional information. */
#define MEMUSAGE1 0.95                  /* Portion of available memory that the program may use if /proc/meminfo has */
                                        /* MemAvailable information. */
#define MEMUSAGE2 0.75                  /* Portion of physical memory that the program may use if /proc/meminfo has not */
                                        /* MemAvailable information. */
#define MEMORYLIMIT1 600000000          /* Minimum memory in bytes for the program to work in normal mode. */
#define SUBSUMTIMEOUT 0.0185            /* Maximum time spent in each check of clause subsumptions. See IsSubsumedByClause() */
                                        /* function global comments for additional information. */
#define SATQMAXTIME 2                   /* Maximum seconds of execution of SatQ2SatSolver() function when type parameter is not zero. */
#define SATQITERS 50                    /* Number of iterations between time checks in SatQ2SatSolver() function. */
#ifdef WINTYPE
#define WPATHSEPARATOR L'\\'            /* Windows type wide char file path separator. */
#else
#define WPATHSEPARATOR L'/'             /* Linux & AIX type wide char file path separator. */
#endif

/* Useful macros */
#ifdef DEBUGMEMORY
#define MYFREE(X) myfree(X,__FILE__,__LINE__)
#define DBGMEMELEMS 2000
#define DBGMEMELEMS 0
#else
#if (MEMCHECK == 1) || (MEMCHECK == 2)
#define MYFREE(X) freemn(X)
#else
#define MYFREE(X) free(X)
#endif
#endif
#if (MEMCHECK == 1) || (MEMCHECK == 2)
#define MYALLOC(X) mallocmn(X)
#define MYREALLOC(X,Y) reallocmn(X,Y)
#define MYSTRDUP(X) strdupmn(X)
#else
#define MYALLOC(X) malloc(X)
#define MYREALLOC(X,Y) realloc(X,Y)
#define MYSTRDUP(X) strdup(X)
#endif
#define ABS(X) ((X)>=0?X:-(X))
#define MAX(X,Y) ((X)>=(Y)?X:Y)
#define MIN(X,Y) ((X)<=(Y)?X:Y)
#define NUMELEMS(X) sizeof(X)/sizeof(X[0])
#define MATCHFLGS11(X,Y) (((X)&ITEM1SEC)==((Y)&ITEM1SEC))
#define MATCHFLGS12(X,Y) (((X)&ITEM1SEC)==(((Y)&ITEM2SEC)>>1))
#define SHORTSYMBOLSIZE sizeof(symbol)-(2*sizeof(void *))
#ifdef MURMUR3
#define ROT32(x, y) ((x << y) | (x >> (32 - y)))
#endif

/* STRUCTURES */
/* Structure for chaining names buffers for told sentences. */
typedef struct snameschain {
	struct snameschain *next;               /* Pointer to first nameschain structure of next older level of names. */
	struct snameschain *peer;               /* Pointer to next nameschain structure of the same level. The new names */
	                                        /* at a level are always added at the end of peers chain. */
	struct snameschain *frstpeer;           /* Pointer to first nameschain structure of the same level. */
	struct snameschain *lstpeer;            /* Pointer to last nameschain structure of the same level. Only used for */
	                                        /* first nameschain structure of each level. */
	struct slkedfitem *quantifier;          /* Pointer to quantifier linked list item corresponding to a variable item. */
	char *name;                             /* Pointer to name of predicate, function or variable. */
	int32_t flags;                          /* PREDICATE for predicates, FUNCTION for functions or 0 for variables. */
	union {
		int32_t arity;                      /* Arity of predicate or function. */
		int32_t varnum;                     /* Number assigned to variable or -1 if variable not yet standardized. */
	} data;
	//struct shaschain *type;                 /* Pointer to type or sort assigned to predicate, function or variable. */
} nameschain;

/* Structure for formula common prefix. */
typedef struct scmprefix {
	struct scmprefix *next;             /* Pointer to next formula for formulas of the same kind (TEXT, ACTIVE, PASSIVE, */
	                                    /* UNPROC y GROUNDJOINABLE) in the KB. NULL if last formula of its kind. */
	struct scmprefix *prev;             /* Pointer to previous formula for formulas of the same kind (TEXT, ACTIVE, PASSIVE, */
	                                    /* UNPROC y GROUNDJOINABLE) in the KB. NULL if first formula of its kind. */
	                                    /* This is intended to free formula memory. */
	struct scmprefix *parent1;          /* First parent formula pointer. */
	struct scmprefix *parent2;          /* Second parent formula pointer, NULL if there is no second parent. */
	                                    /* For predicate definition formulas this pointer is used to chain */
	                                    /* the set of predicate definition formulas used in a single formula */
	                                    /* renaming inference. */
	struct scmprefix *prevdisp;         /* Pointer to previous formula in disposal queue. NULL if last formula */
	                                    /* or formula not in disposal queue. This queue interconnects clauses */
	                                    /* from other queues.*/
	union {
		char *text;                     /* Pointer to formula in plain text form. Used for text type formulas. */
		struct sbinprefix *bin;         /* Pointer to binary clause prefix structure. Used for compiled clauses. */
	} part2;
	uint64_t number;                    /* Number assigned to formula. */
	int32_t inference;                  /* Inference from where the formula comes from (see definitions above). */
	uint32_t flags;                     /* See flags above. This must be the last field of this structure. */
} cmprefix;

/* Structure for binary clause prefix. */
typedef struct sbinprefix {
	cmprefix *clause;                     /* Pointer to cmprefix structure of the formula itself. */
	cmprefix *nextselect[2*SELECTQUEUES]; /* Set of pointers to next formula in each clause selection queue. NULL if last formula. */
	cmprefix *prevselect[2*SELECTQUEUES]; /* Set of pointers to previous formula in each clause selection queue. NULL if last formula. */
	union {
		uint8_t *asserts;                 /* Pointer to buffer with clause assertions. */
		uint8_t *cptopterm;               /* Pointer to buffer critical pair top term. */
	} ovly;
	uint8_t *lockasserts;                 /* Pointer to buffer with clause locking assertions. */
	uint64_t signature;                   /* Signature of last inference process. */
	uint64_t locksignature;               /* Signature for lock and unlock processes. */
	uint32_t literals;                    /* Number of literals in clause. */
	uint32_t clweight;                    /* Clause weight. */
	int32_t size;                         /* Size in bytes of formula including ending UNITEND byte. */
	int32_t asize;                        /* Size in bytes of assertions including ending UNITEND byte. */
	int32_t lasize;                       /* Size in bytes of locking assertions assertions including ending UNITEND byte. */
	int32_t maxvarnb;                     /* Maximum variable number, -1 if no variables. This is useful for */
	                                      /* memory allocations in ordering functions. */
	int32_t equflg;                       /* 0 if equality ordering data has not been set in equalities. */
	                                      /* 1 if equality ordering data has been set in SELECTED equalities */
	                                      /* by calls from SelectItems() function. */
	                                      /* 2 if equality ordering data has been set in the only equality */
	                                      /* of a clause used for backward demodulation. */
	                                      /* See comments in FreeEquData() function for indications on how */
	                                      /* equality ordering data is allocated and freed with the help of this flag. */
	uint32_t sinedist;                    /* SInE distance of the clause to the conjecture for clause selection. */
	uint32_t agedist;                     /* Age distance for clause selection. */
	int32_t avdist;                       /* Avatar distance for clause selection. */
	int32_t hrndist;                      /* Horn distance for clause selection. */
	uint8_t formula[1];                   /* Dummy for start of formula items. This must be the last field of this structure. */
} binprefix;

#ifdef SEMANTICTAUTOLOGY
/* Structure for binary tautology clause. */
typedef struct sttlgclause {
	struct sttlgclause *next;           /* Pointer to next clause for (in)equalities of the same kind (positive or negative), */
	                                    /* NULL if last clause of its type. */
	struct sttlgclause *prev;           /* Pointer to previous clause for (in)equalities of the same kind (positive or negative), */
                                        /* NULL if first clause of its type. */
	uint8_t formula[1];                 /* Dummy for start of formula items. This must be the last field of this structure. */
} ttlgclause;
#endif

/* Structure for SAT clauses. */
typedef struct ssatclause {
	struct ssatclause *glblnext;        /* Pointer to next SAT clause of the same global queue, NULL if last. */
	struct ssatclause *glblprev;        /* Pointer to previous SAT clause of the same global queue, NULL if last. */
	struct ssatclause *next;            /* Pointer to next SAT clause of the SAT solver ordered clauses queue, NULL if last. */
	struct ssatclause *prev;            /* Pointer to previous SAT clause of the SAT solver ordered clauses queue, NULL if last. */
	struct scmprefix *parent;           /* Pointer to parent A-clause. */
	uint64_t number;                    /* Number assigned to the SAT clause. */
	int32_t inference;                  /* Inference from which the SAT clause was obtained. It can be SPLIT or CONTRADICTION. */
	                                    /* See definitions above. */
	int32_t literals;                   /* Number of literals in clause. */
	int32_t satlits;                    /* Number of satisfied literals in clause. */
	int32_t undeflits;                  /* Number of undefined literals in clause. */
	int32_t inproof;                    /* Set to 1 if clause is in the unsatisfiable core but is not the contradiction clause */
	                                    /* that caused the SAT refutation, set to 2 if clause is in the unsatisfiable core and */
	                                    /* it is the contradiction clause that caused the SAT refutation, zero otherwise. */
	uint8_t formula[1];                 /* Dummy for start of formula items. This must be the last field of this structure. */
} satclause;

/* Structure for statistic information. */
/* If new fields are added to this structure the following changes must be made */
/* to the involved functions: */
/* - Search for an existing field and add the appropriate equivalent code */
/*   for the new field. */
/* - Update totals calculation in PrintStatData() function if appropriate. */
/* - Add code for printing proof lines in PrintProof() function if appropriate. */
typedef struct sstats {
	uint64_t tautologies;               /* Number of syntactic tautologies detected. */
	#ifdef SEMANTICTAUTOLOGY
	uint64_t semtautologies;            /* Number of semantic tautologies detected. */
	#endif
	uint64_t fwdsubsum;                 /* Number of forward subsumptions performed. */
	uint64_t bcksubsum;                 /* Number of backward subsumptions performed. */
	uint64_t factors;                   /* Number of factoring inferences. */
	uint64_t removedups;                /* Number of duplicate literal removing simplifications. */
	uint64_t fwdemodulations;           /* Number of forward demodulation inferences. */
	uint64_t bkdemodulations;           /* Number of backward demodulation inferences. */
	uint64_t fwsubsresltns;             /* Number of forward subsumtion resolution inferences. */
	uint64_t bksubsresltns;             /* Number of forward subsumtion resolution inferences. */
	uint64_t triveqres;                 /* Number of trivial equality resolution inferences. */
	uint64_t desteqres;                 /* Number of destructive equality resolution inferences. */
	uint64_t resolutions;               /* Number of resolution inferences. */
	uint64_t paramodulations;           /* Number of paramodulation inferences. */
	uint64_t equresolutions;            /* Number of equality resolution inferences. */
	uint64_t equfactorings;             /* Number of equality factoring inferences. */
	uint64_t splits;                    /* Number of split inferences. */
	uint64_t eqsplits;                  /* Number of equality splits inferences. */
	uint64_t prennfxform;               /* Number of pre-nnf inferences. */
	uint64_t nnfxform;                  /* Number of nnf inferences. */
	uint64_t miniscoping;               /* Number of miniscoping inferences. */
	uint64_t skolemize;                 /* Number of skolemization inferences. */
	uint64_t predicatedef;              /* Number of predicate definitions for sub-formula renaming. */
	uint64_t formularenaming;           /* Number of formula renaming inferences. */
	uint64_t cnfconversion;             /* Number of CNF conversion inferences. */
	uint64_t tfsimplif;                 /* Number of $true and $false simplifications. */
	uint64_t sattautologies;            /* Number of tautologies detected in SAT clauses. */
	uint64_t satremovedups;             /* Number of duplicate literal removing simplifications in SAT clauses. */
	uint64_t satsubsum;                 /* Number of subsumptions performed for SAT clauses. */
	uint64_t nbactive;                  /* Number of active clauses. */
	uint64_t nbpassive;                 /* Number of passive clauses. */
	uint64_t nbunproc;                  /* Number of unprocessed clauses. */
	uint64_t nbgrjnbl;                  /* Number of ground joinable clauses. */
	uint64_t memory;                    /* Memory used including non returnable unused space assigned to the program. */
	uint64_t netmemory;                 /* Memory used excluding non returnable unused space assigned to the program. */
	double elapsed_time;                /* Elapsed time in seconds. */
} stats;

/* Structure for program options. */
typedef struct soptions {
	int32_t fweight;                    /* Type of function weight (UNIFORM or ARITY). */
	int32_t termord;                    /* Terms ordering type (STANDARD or NONRECURSIVE). */
	int32_t litord;                     /* Literals ordering type (STANDARD, NONRECURSIVE or LEXICOGRAPHIC). */
	int32_t select;                     /* Type of literal selection (see defines above). */
	int32_t demodulation;               /* Demodulation options (DEMODON, DEMODOFF, COMPLETE). */
	int32_t factoring;                  /* Factoring is enabled if this is not zero. */
	int32_t algorithm;                  /* Selected algorithm (OTTER or DISCOUNT). */
	int32_t mxwghtopt;                  /* MXWEIGHTxxx (see definitions above). */
	uint32_t maxweight;                 /* Maximum weight allowed. */
	int32_t layer;                      /* Type of specialized layers in use by current strategy, see defines above. */
	int32_t layerset[2];                /* Types of specialized layers allowed for current problem, see defines above. */
	int32_t lookahead;                  /* Use lookahead to modify weight queues for clause clause selection. */
	int32_t split;                      /* Splitting and interface between first order prover and */
	                                    /* SAT solver is enabled if this is not zero. */
	int32_t shuffle;                    /* If not zero then include strategies with shuffling and randomization. */
	int32_t grjoin;                     /* Ground joinability setting when UEQDSC saturation algorithm is used. */
	                                    /* 0 -> OFF, 1 -> ON, 2 -> MED. */
	int32_t connect;                    /* Connectedness setting when UEQDSC saturation algorithm is used, 0 -> OFF, 1 -> ON. */
	int32_t weakrw;                     /* Weak rewrite setting for UEQ problems, 0 -> OFF, 1 -> ON. */
	float timeout;                      /* Timeout setting. */
	float gamma;                        /* Factor that multiplies a non learning weight for learning weight calculation. */
} options;

/* Structure for hash separate symbol chaining. */
typedef struct shashchain {
	struct shashchain *nextelem;        /* Pointer to next hashchain element or NULL if last. This pointer is */
	                                    /* used for freeing memory allocated to hashchain elements. */
	struct shashchain *nextsymbol;      /* Pointer to next hashchain element of the same hash value */
	                                    /* or NULL if last. */
	void *discr[DSCTREETYPES];          /* Set of pointers to discrimination trees. See EditDiscrTrees() */
	                                    /* global function comments for an index description. */
	#ifdef SEMANTICTAUTOLOGY
	void *posdsctrttlg;                 /* Pointer to semantic tautology discrimination tree for functions and positive predicates */
	                                    /* and equalities. */
	void *negdsctrttlg;                 /* Pointer to semantic tautology discrimination tree for negative predicates equalities. */
	#endif
	#ifdef DEBUGTREE
	uint32_t numleafs[DSCTREETYPES];    /* Number of leafs of each discrimination tree. */
	#endif
	int32_t arity;                      /* Symbol arity. */
	int32_t type;                       /* Type flags, see flag definitions above. */
	int32_t precedence;                 /* Symbol precedence value for ordering purposes. */
	int32_t precedencebk;               /* Copy of precedence field to restore after randomization. */
	int32_t usecount;                   /* Number of times the symbol was used in the initial clausification of a problem. This is used for */
	                                    /* SInE distance calculation. If PRECEDENCE is 1 it is also used for for symbol ordering precedence. */
	uint32_t sinedist;                  /* SInE distance of the symbol to the conjecture. */
	uint32_t hash;                      /* Symbol 32 bits hash value corresponding to symbol name. */
} hashchain;

/* Structure for hash component symbol chaining. */
typedef struct shashchcmp {
	struct shashchcmp *nextelem;        /* Pointer to next hashchcmp element or NULL if last. This pointer is */
	                                    /* used for freeing memory allocated to hashchcmp elements. */
	struct shashchcmp *nextsymbol;      /* Pointer to next hashchcmp element of the same hash value */
	                                    /* or NULL if last. */
	struct shashchcmp *nextundef;       /* Pointer to next undefined hashchcmp element of the same component. */
	struct shashchcmp *prevundef;       /* Pointer to previous undefined hashchcmp element of the same component. */
	struct sasymbol *frstposassrt;      /* Pointer to first positive asymbol in assertions of first order prover. */
	struct sasymbol *frstnegassrt;      /* Pointer to first negative asymbol in assertions of first order prover. */
	struct sasymbol *firstlock;         /* Pointer to first asymbol in locks of first order prover. */
	struct sasymbol *frstsat;           /* Pointer to first asymbol in SAT solver. */
	struct ssatnode *satnode;           /* Pointer to satnode that assigned a value to this component. */
	uint8_t *formula;                   /* Pointer to clause binary formula of component clause associated */
	                                    /* with the P-predicate. */
	int32_t size;                       /* Size in bytes of formula including ending UNITEND byte. */
	int32_t literals;                   /* Number of literals in formula. */
	int32_t number;                     /* Number assigned to symbol name, the xxx in spl_xxx name. */
	uint64_t defnumber;                 /* Number assigned to the symbol definition in the proof. The component */
	                                    /* clause is assigned the number defnumber+1. */
	uint32_t hash;                      /* Symbol 32 bits hash value corresponding to clause component. */
	uint32_t flags;                     /* Flags for this component, see definitions above. */
	uint32_t npositive;                 /* Number of clauses in SAT solver with a positive literal for this component. */
	uint32_t nnegative;                 /* Number of clauses in SAT solver with a negative literal for this component. */
	uint32_t absdif;                    /* Absolute value of npositive field minus nnegative field. */
	uint32_t sinedist;                  /* SInE distance of the clause that originated the component to the conjecture. */
	int32_t agedist;                    /* Age distance of the clause that originated the component to the conjecture. */
} hashchcmp;

/* Structure for coefficients of weight ordering polynomials. */
typedef struct scoeff {
	int32_t alpha0;                     /* First coefficient of weight polynomial. */
	int32_t *alpha;                     /* Pointer to second and subsequent coefficients of weight polynomial. */
	struct scoeff *eqterm1;             /* Pointer to coeff structure of first root term in an equality. */
	                                    /* NULL if current structure is not for an equality literal. */
	struct scoeff *eqterm2;             /* Pointer to coeff structure of second root term in an equality. */
	                                    /* NULL if current structure is not for an equality literal. */
} coeff;

/* Structure for coefficients of weight ordering polynomials, compact format. */
/* These are used for equality data saved in long term storage. */
typedef struct scoeffcpt {
	int32_t coefnb;                     /* Coefficient number. Number 0 corresponds to symbols coefficient. */
	                                    /* Number n corresponds to variable n-1 for n>1. A negative value */
	                                    /* indicates the end of the list. */
	int32_t coef;                       /* Coefficient value. Not used in the last list element. */
} coeffcpt;

/* Structure for coefficients of weight ordering polynomials, compact format and linked list. */
/* These are used for polynomials saved in short term storage. The linked list is ordered */
/* for increasing values of the coefnb field. */
typedef struct slkedcoef {
	int32_t coefnb;                     /* Coefficient number. Number 0 corresponds to symbols coefficient. */
	                                    /* Number n corresponds to variable n-1 for n>1. */
	int32_t coef;                       /* Coefficient value. */
	struct slkedcoef *next;             /* Pointer to next polynomial term, NULL if last term. */
} lkedcoef;

/* Structure for linear polynomial node of tree structure for evaluating */
/* ordering constraints of equality root terms. */
typedef struct sequpnode {
	coeffcpt *coefs;                    /* Pointer to linear coefficients of term in compact form. */
	int32_t result;                     /* Indicates the result (EQUAL, GREATER, LESSER or FAILURE) of an execution */
	                                    /* of CompareEquTerms(), that is, with no substitution applied. It also */
	                                    /* indicates the result of the call to CompareItemWeight2() done */
	                                    /* from CompareEquTerms() with  EQUALW, NOGREATER, NOLESSER */
	                                    /* or FAILUREW flags. */
	struct sequpnode *subterms[1];      /* Set of pointers to child (subterms) nodes. This must be the last item of this structure. */
} equpnode;

/* Structure for symbols (literals and functions) items data. */
typedef struct ssymbol {
	int32_t offset;                     /* Offset from beginning of formula field buffer to start of current item. */
	int32_t nextoff;                    /* Offset from beginning of formula field buffer to next item */
	                                    /* that is not a subterm of current item. */
	hashchain *symbol;                  /* Pointer to hashchain structure for the item symbol. */
	equpnode *leftnode;                 /* Pointer to left node equpnode structure. */
	equpnode *rightnode;                /* Pointer to right node equpnode structure. */
} symbol;

/* Structure for assertion symbols (P-predicates, component names). */
typedef struct sasymbol {
	hashchcmp *symbol;                  /* Pointer to hashchain structure for the item symbol. */
	struct sasymbol *next;              /* Pointer to next asymbol structure for the same component of the same class. */
	                                    /* (assertion, lock or SAT clause literal). */
	struct sasymbol *prev;              /* Pointer to previous asymbol structure for the same component of the same class. */
	union {
		binprefix *binp;                /* Pointer to A-clause binary formula to which the asymbol belongs if asymbol */
		                                /* belongs to assertions or lock assertions of an A-clause. */
		satclause *satcl;               /* Pointer to SAT clause binary formula to which the asymbol belongs if asymbol */
		                                /* belongs to a SAT clause. */
	} part2;
} asymbol;

/* Discrimination tree node structure. */
typedef struct sdstrnode {
	struct sdstrnode *parent;           /* Pointer to parent node, NULL if first level node. The discr[] field of the */
	                                    /* symbols hashchain structure points to a first level node or a leaf. */
	struct sdstrnode *nxtpeer;          /* Pointer to next peer node. */
	struct sdstrnode *prevpeer;         /* Pointer to previous peer node. */
	union {
		struct sdstrnode *childnode;    /* Pointer to first child node. */
		struct sdstrleaf *childleaf;    /* Pointer to first child leaf. */
	} child;
	hashchain *symbol;                  /* Pointer to node symbol for nodes that are not for variables. Not used */
	                                    /* for nodes that are variables. */
	uint8_t *item;                      /* This is used for temporary storage. DsTermIndexing() function uses it */
	                                    /* to store the pointer to the next clause term that is not a sub-term */
	                                    /* of the clause item matching the node if it has sub-terms or NULL otherwise. */
	                                    /* This is used later to update pndbrkts field. QueryDscTree() function uses it */
	                                    /* to store the pointer to clause item that matches the node or NULL if this is */
	                                    /* a sub-term of a node item that matches a variable. This is used later for */
	                                    /* backtracking. See DsTermIndexing() */
	                                    /* global comments. */
	uint16_t pndbrkts;                  /* Number of brackets to be closed just after the node symbol, see DsTermIndexing() */
	                                    /* global comments. */
	uint8_t nodetype;                   /* VARIABLE if node is for a variable, zero otherwise. */
	int8_t childtype;                   /* Type of child, it can be NODECHILD or LEAFCHILD (see definitions above). */
} dstrnode;

/* Discrimination tree leaf structure. */
typedef struct sdstrleaf {
	struct sdstrnode *parent;           /* Pointer to parent node. */
	struct sdstrleaf *peer;             /* Pointer to next peer (same parent) leaf. */
	#ifdef TREELEAFSFIFO
	struct sdstrleaf *lstpeer;          /* Pointer to last peer leaf in the chain. Only used by first leaf in the chain. */
	#endif
	uint8_t *item;                      /* Pointer to the tree indexed clause item. */
} dstrleaf;

#ifdef SEMANTICTAUTOLOGY
/* Discrimination tree node structure. */
typedef struct sttlgdstrnode {
	struct sttlgdstrnode *parent;        /* Pointer to parent node, NULL if first level node. The discr[] field of the */
	                                     /* symbols hashchain structure points to a first level node or a leaf. */
	struct sttlgdstrnode *nxtpeer;       /* Pointer to next peer node. */
	struct sttlgdstrnode *prevpeer;      /* Pointer to previous peer node. */
	union {
		struct sttlgdstrnode *childnode; /* Pointer to first child node. */
		struct sttlgdstrleaf *childleaf; /* Pointer to first child leaf. */
	} child;
	union {
		hashchain *symbol;               /* Pointer to node symbol for nodes that are not for variables. Not used */
		                                 /* for nodes that are variables. */
		uint16_t varnb;                  /* Variable number. */
	} itemid;
	uint8_t nodetype;                    /* VARIABLE if node is for a variable, zero otherwise. */
	int8_t childtype;                    /* Type of child, it can be NODECHILD or LEAFCHILD (see definitions above). */
} ttlgdstrnode;

/* Discrimination tree leaf structure. */
typedef struct sttlgdstrleaf {
	struct sttlgdstrnode *parent;        /* Pointer to parent node. */
	struct sttlgdstrleaf *peer;          /* Pointer to next peer (same parent) leaf. */
	uint8_t *item;                       /* Pointer to the tree indexed item. */
	uint8_t *formula;                    /* Pointer to the formula containing the tree indexed item. */
} ttlgdstrleaf;
#endif

/* Discrimination tree backtrack structure for QueryDscTree() function. */
typedef struct sdstrbktrk {
	dstrleaf *lstmatch;                 /* Pointer to last match found in a previous call. */
	uint8_t **treeitems;                /* Pointer to set of tree item pointers. */
	uint32_t nodeidx;                   /* Index of current node in treeitems field set. */
} dstrbktrk;

/* Structure for indexing of positive equalities selected for paramodulating from them */
/* in active clauses that have maximal root terms that are variables. */
typedef struct sdsendnode {
	int32_t size;                       /* Number of total symbol pointers allocated. */
	int32_t used;                       /* Number of used symbol pointers. */
	uint8_t **symbol;                   /* Pointer to set of pointers to item in the clause that matches the tree leave. */
} dsendnode;

/* Structure for SAT solver node. */
typedef struct ssatnode {
	struct ssatnode *nextnode;          /* Pointer to next SAT solver node, NULL if last.*/
	struct ssatnode *prevnode;          /* Pointer to previous SAT solver node, NULL if first.*/
	hashchcmp *symbol;                  /* SAT symbol that was assigned a value true or false in this node. */
	int32_t flags;                      /* Flags for this node: POSITIVE or NEGATIVE and possibly LASTOPTION. */
	satclause *forcedcl;                /* Pointer to unit clause that forced the value of the node symbol field, */
	                                    /* NULL if the node was not forced by a unit clause. */
	uint32_t number;                    /* Sequence number for this node. */
} satnode;

/* Structure for proof node. */
typedef struct sproofnode {
	struct sproofnode *nextnode;        /* Pointer to next proof node, NULL if last.*/
	struct sproofnode *prevnode;        /* Pointer to previous proof node, NULL if first.*/
	struct sproofnode *backptr;         /* This is used by AddAclause2Proof() for backtracking. */
	int32_t type;                       /* Type of node, see definitions above. */
	union {
		hashchcmp *symbol;              /* Pointer to component symbol. */
		satclause *satclause;           /* Pointer to SAT clause. */
		cmprefix *Aclause;              /* Pointer to A-clause. */
	} ptr;
} proofnode;

/* Structure for vertical features scan node. */
typedef struct svertnode {
	struct svertnode *nextnode;         /* Pointer to next node, NULL if last.*/
	struct svertnode *prevnode;         /* Pointer to previous node, NULL if first.*/
	uint8_t *item;                      /* Pointer to clause item. */
	uint8_t *nextitem;                  /* Next item jumping over subterms of current item. */
	int32_t depth;                      /* Node depth, starting with zero. */
} vertnode;

/* Sparse learning feature vector structure. See GetClauseVector() function for a description */
/* of each feature in vector. */
typedef struct slearnvector {
	int32_t size;                       /* Number of nun null elements in sparse vector. */
	int32_t *index;                     /* Pointer to set of non null vector component indices. */
	double *sparsevect;                 /* Pointer to non null vector components. */
} learnvector;

/* Structure for learning record. */
typedef struct slearnrecord {
	uint64_t header;                    /* Record header with flags indicating the relevant part of the strategy */
	                                    /* to which the learn record applies. The relevant strategy flags are those */
	                                    /* corresponding to ALGMASK, SELECTMASK, TERMORDRMASK, LITORDRMASK and */
	                                    /* FUNCWMASK masks .The most significant byte of the header contains the */
	                                    /* problem type, a value between 0 and PBMTYPES-1. */
	learnvector vector;                 /* Sparse learning feature vector structure. */
} learnrecord;

/* Structure for KB definition. */
typedef struct skbase {
	cmprefix *firsttxt;                 /* Pointer to first text formula, NULL if no text formulas added yet. */
	                                    /* Older formulas are placed first and this pointer points to the oldest */
	                                    /* text formula. */
	cmprefix *lasttxt;                  /* Pointer to last text formula, NULL if no text formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* text formula. */
	cmprefix *frstpassive;              /* Pointer to first passive formula, NULL if no passive formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* passive formula. */
	cmprefix *lstpassive;               /* Pointer to last passive formula, NULL if no passive formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* passive formula. */
	cmprefix *frstactive;               /* Pointer to first active formula, NULL if no active formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* active formula. */
	cmprefix *lstactive;                /* Pointer to last active formula, NULL if no active formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* active formula. */
	cmprefix *frstunproc;               /* Pointer to first unprocessed formula, NULL if no unprocessed formulas */
                                        /* added yet. Older formulas are placed first and this pointer points */
                                        /* to the oldest unprocessed formula. */
	cmprefix *lstunproc;                /* Pointer to last unprocessed formula, NULL if no unprocessed formulas */
                                        /* added yet. Newer formulas are placed last and this pointer points */
                                        /* to the newest unprocessed formula. */
	cmprefix *frstgrjoin;               /* Pointer to first ground joinable formula, NULL if no ground joinable */
                                        /* formulas added yet. Older formulas are placed first and this pointer */
                                        /* points to the oldest ground joinable formula. */
	cmprefix *lstgrjoin;                /* Pointer to last ground joinable formula, NULL if no ground joinable */
                                        /* formulas added yet. Newer formulas are placed last and this pointer */
                                        /* points to the newest ground joinable formula. */
	cmprefix *frstlocked;               /* Pointer to first locked formula, NULL if no locked formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* locked formula. */
	cmprefix *lstlocked;                /* Pointer to last locked formula, NULL if no locked formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* locked formula. */
	cmprefix *lastdisp;                 /* Pointer to last formula in disposal queue, NULL if no formulas */
	                                    /* added yet. This queue interconnects clauses from other queues. */
	cmprefix *negeq;                    /* Last negative equality added to UNPROC or PASSIVE clauses when saturation */
	                                    /* algorithm is UEQDscnt(). */
	proofnode *frstprfnode;             /* Pointer to first proof node. */
	satclause *firstsatq;               /* Pointer to first SAT clause in the SAT queue. */
	satclause *lastsatq;                /* Pointer to last SAT clause in the SAT queue. */
	satclause *frstsatglbl;             /* Pointer to first SAT clause in the SAT solver global queue. */
	satclause *lastsatglbl;             /* Pointer to last SAT clause in the SAT solver global queue. */
	satclause *frstunsats;              /* Pointer to first SAT clause in the SAT solver ordered clause queue. */
	satclause *lstunsats;               /* Pointer to last SAT clause in the SAT solver ordered clause queue. */
	satclause *frstunsats2;             /* Pointer to start of the part of SAT solver ordered clause queue with clauses */
	                                    /* that are unsatisfied and have two or more undefined literals. */
	hashchcmp *frsthshundf;             /* Pointer to first undefined hashchcmp element for this KB in the hashchcmp ordered queue. */
	hashchcmp *frsthunbp;               /* Pointer to first undefined hashchcmp element with clauses with negative and */
	                                    /* positive literals (both polarities) for this KB in the hashchcmp ordered queue. */
	hashchcmp *lasthshundf;             /* Pointer to last undefined hashchcmp element for this KB in the hashchcmp ordered queue. */
	satnode *satrootnode;               /* SAT solver root node pointer. */
	learnvector *lrnvector;             /* Pointer to sparse learning vector structure set or NULL if no learning vectors are available. */
	dsendnode vartrmidx;                /* Indexes of positive equalities selected for paramodulating from them in active clauses */
	                                    /* that have maximal root terms that are variables. */
	uint64_t signature;                 /* Last signature used for inference process. */
	uint64_t locksignature;             /* Last signature used for lock and unlock processes. */
	int32_t flags;                      /* KB flags, see definitions above. */
	uint64_t nformulas;                 /* Number of formulas in KB so far. */
	options opts;                       /* Options for KB. */
	int32_t selectcnt;                  /* Number of extractions done so far in the current round-robin clause extraction cycle. */
	int32_t lrscnt;                     /* Count of clauses removed from the passive queue in the current LRS cycle. */
	uint32_t splsymbols;                /* Number of P-predicates for P-predicate name assignment. */
	int32_t status;                     /* KB status (see flags above). */
	struct timeval time;                /* Start time of current saturation algorithm. */
	struct timeval endtime;             /* End time of current saturation algorithm. */
	stats prstats;                      /* KB statistics related to this KB. This depends on the context (process, main or working KB). */
} kbase;

/* Structure for substitutions. See Unify() function for substitution format. */
typedef struct ssubst {
	uint8_t *buffer;                    /* Pointer to buffer with substitution. May be NULL. */
	                                    /* See Unify() function for format description. */
	uint32_t buffsize;                  /* Total buffer size in bytes. */
	uint32_t substsize;                 /* Size in bytes of substitutions including the UNITEND byte. */
} subst;

/* Structure to store subsuming related data for each literal in a subsuming or subsumed formula. */
typedef struct ssubsumdata {
	int32_t lits1;                      /* Number of literals in candidate subsumed formula. */
	int32_t lits2;                      /* Number of literals in candidate subsuming clause. */
	uint8_t ***pinst;                   /* Address of set of pointers, one pointer for each literal in the candidate */
	                                    /* subsuming clause. Each pointer is the address of the set of pointers to */
	                                    /* instance literals in candidate subsumed formula for each candidate subsuming */
	                                    /* literal. */
	int32_t **instnb;                   /* Address of set of pointers, one pointer for each literal in the candidate */
                                        /* subsuming clause. Each pointer is the address of the set of instance literal */
	                                    /* numbers in candidate subsumed formula for each candidate subsuming literal. */
	int32_t *nbofinst;                  /* Address of set containing the number of instance literals for each literal */
	                                    /* in the subsuming clause. */
	int32_t *usedinst;                  /* Address of set containing one element for each literals in candidate subsumed */
	                                    /* clause that is set to 1 if the literal has been already assigned to a literal */
	                                    /* in the candidate subsuming clause or 0 otherwise. */
} subsumdata;

/* Structure for linked list formula format. */
typedef struct slkedfitem {
	struct slkedfitem *next;            /* Pointer to next linked list item. */
	struct slkedfitem *prev;            /* Pointer to previous linked list item. */
	union {
		struct slkedfitem *mtchbrckt;   /* Pointer to matching bracket item. */
		nameschain *vardata;            /* Pointer to nameschain structure of a quantifier or variable item. */
		hashchain *symbol;              /* Pointer to hashchain structure of a predicate or function item. */
	} ptr;
	union {
		hashchain *sksymbol;            /* Pointer to hashchain skolem symbol in existential quantifier or variable items. */
		int16_t varnumber;              /* Number assigned to standard name of variable in a quantifier or variable item. */
		struct {
			int32_t posclauses;         /* Lower bound of number of clauses generated for a block for the */
			                            /* open bracket or literal starting the block. */
			int32_t negclauses;         /* Lower bound of number of clauses generated for the negation of a block*/
			                            /* for the open bracket or literal starting the block. */
		} clauses;
	} data;
	struct slkedfitem *dupq;            /* Auxiliary pointer to duplicate quantifier, used when duplicating blocks. */
	int32_t flags;                      /* Item flags. */
} lkedfitem;

/* Structure for multi-processing control. */
typedef struct smpctl {
	clock_t start;                      /* CPU ticks at start of saturation process. */
	clock_t end;                        /* CPU ticks at end of saturation process. */
	int32_t status;                     /* Process status. See definitions above. */
	int32_t prfprntinproc;              /* Set to non zero if proof printing is in process, zero otherwise. */
	int32_t solvekb;                    /* Process number for problems solved in the saturation stage (see procnb variable), */
	                                    /* active_cores for problems solved in the pre-process, -1 if problem was not solved. */
	sem_t procctl;                      /* Semaphore for process status updates. */
	sem_t statistics;                   /* Semaphore for statistics update. */
	sem_t strategy;                     /* Semaphore for process strategy selection. */
} mpctl;

/*----------------------------------------------------------------
 *
 *  Global variables shared with other modules that are initialized here.
 *  WARNING: Don't use "static" as then a different variable will be created in
 *  each module with the same name.
 *
 *--------------------------------------------------------------*/
#ifndef VARTYPE
#define VARTYPE extern
extern double tot_cpu_time; /* Total CPU time used in seconds. */
extern int32_t tot_pbms; /* Total number of problems examined. */
extern int32_t solved_pbms; /* Total number of problems solved. */
extern int32_t queuerts[2]; /* Ratio numbers for weight queues and age queue. */
extern int32_t *slctordr; /* Pointer to clause selection order queue. The I-th clause selection is picking the clause */
                          /* from the slctordr[I] clause selection queue. */
extern hashchain *kb_hash; /* First chained hashchain element. */
extern hashchcmp *kb_hashcmp; /* First chained hashchcmp element. */
extern int32_t syntaxmode; /* Syntax type: TPTP or DRODI. */
extern int32_t proofmode; /* Proof printing type: TPTP or DRODI. */
extern int32_t prtkbdata; /* PRINTKBDATA type: 0->OFF, 1->ON */
extern int32_t verbose; /* Set to 2 if VERBOSE option is ALL, 1 if it is ON, zero otherwise. */
extern char *importfile; /* Pointer to IMPORT file name or NULL. */
extern char *learnto; /* Pointer to LEARNTO file name or NULL. */
extern char *learnfrom; /* Pointer to LEARNFROM file name or NULL. */
extern learnrecord *lrnrecords; /* Learning records with headers and vectors from learning file. */
extern int32_t learnnew; /* If zero then learn-to output will be appended to learn-to file. Otherwise a new file */
                         /* will be created. */
extern int32_t loadredundant; /* If not zero then add formulas with lemma, theorem and corollary roles, */
                       /* otherwise these formulas are not loaded. */
extern int32_t timeslices; /* Number of time slices. */
extern double ppthrshld; /* Threshold for the preprocessing clause relevance filtering. */
                  /* See Preprocess() global comments for details. */
extern int32_t prproflags; /* Preprocessing clause filtering option flags. This can be CLFILTERINGOFF, CLFILTERINGON, */
                    /* CLFILTERINGNONEQU and/or CLFILTERINGRESIGN (see definitions above). */
extern int16_t pbmtypestrats[2][PBMTYPES][TUNEDSTRATEGIES]; /* Strategies to use for each type of problem. */
                                     /* See InitStrategies() for details. */
extern uint32_t usrparam; /* User parameter options. */
extern uint32_t usrparam_mask; /* User parameter options mask. */
extern volatile sig_atomic_t signalflag; /* Not zero if a program error caught signal process is in progress. */
#if (MEMCHECK == 1) || (MEMCHECK == 2)
extern uint64_t allcdmemory; /* Currently allocated memory. */
#endif
extern pid_t *pids; /* Pointer to set of child process pids. */
extern uint64_t memorylimit; /* User specified memory limit. */
extern uint64_t initavmemory; /* Available memory at the start of the program. */
extern void *mmapptr; /* Pointer to mmap shared memory. */
extern char *logpath; /* Pointer to path for exceptions logging. */
extern int32_t *fvidx; /* Pointer to non null components of feature vector generated by GetClauseVector() function. */
extern uint32_t usrseed; /* User provided base seed for pseudo random numbers. */
extern uint32_t seeds[PBMTYPES]; /* Default base seeds for each problem type. */
#if defined(DEBUGCODE) && defined(VERBOSE)
extern int32_t verbfrom; /* Initial clause number for verbose tracking. */
extern int32_t verbto; /* Final clause number for verbose tracking. */
#endif
#else
double tot_cpu_time=0.0;
int32_t tot_pbms=0;
int32_t solved_pbms=0;
int32_t queuerts[2]={5,1};
int32_t *slctordr=NULL;
hashchain *kb_hash=NULL;
hashchcmp *kb_hashcmp=NULL;
int32_t loadredundant=1;
int32_t syntaxmode=TPTP;
int32_t proofmode=TPTP;
int32_t prtkbdata=0;
int32_t verbose=0;
char *importfile=NULL;
char *learnto=NULL;
int32_t learnnew=0;
char *learnfrom=NULL;
learnrecord *lrnrecords=NULL;
int32_t timeslices=2;
double ppthrshld=CLFILTERINGTIMEOUT;
int32_t prproflags=CLFILTERINGNONEQU;
int16_t pbmtypestrats[2][PBMTYPES][TUNEDSTRATEGIES]=
{{{101,1015,633,94,1177,1575,335,280,103,631,355,818,1374,1355,533,2032,822,1196,1729,1319,1332,1973,359,1009,1921,1233,996,1815,1608,821,295,1353,
		274,457,823,1415,100,1935,1351,1075,98,635,1253,654,1672,462,476,2089,239,612,1613,1259,289,513,962,816,1455,1878,281,835,154,1173,157,700,
		1755,299,820,1002,343,1433,278,455,451,1194,2122,836,1549,1066,599,719,432,1561,1003,276,153,1369,1995,1968,1001,115,874,453,877,1420,2115,
		1019,1000,1182,713,1063,998,815,1171,1762,116,474,886,959,152,938,1831,283,96,252,873,1635,649,1518,1967,834,275,1373,1361,282,1375,517,460,
		166,155,1377,831,1579,91,450,1571,1187,1676,163,779,2037,2143,1876,1296,1938,1376,1877,30,995,1013,641,655,1180,1237,1849,1054,811,1523,637,
		883,2025,1362,1931,467,515,118,59,1767,1693,512,1298,656,1578,750,1517,95,293,881,1489,160,463,1181,119,699,694,1883,810,523,271,102,640,
		1751,475,813,2036,1677,419,1077,1516,218,1783},
{973,93,1560,961,249,253,813,960,241,969,73,993,1920,781,789,793,273,1470,61,69,823,780,247,242,991,453,1013,103,60,67,962,271,633,833,1003,1830,967,
		782,91,1353,283,240,787,62,811,1173,835,786,1935,83,1015,246,1845,1323,295,966,1575,1143,265,1491,248,1485,783,85,1851,68,270,963,985,1941,
		788,90,423,805,1581,968,990,603,115,66,1934,243,822,1476,1339,810,63,102,792,619,1844,282,1566,441,1574,1002,72,621,1484,252,309,1650,1921,
		1926,129,1740,1561,972,1029,2100,1471,1836,849,1679,1930,279,1831,1923,989,1875,999,1568,1833,269,1605,819,1931,1473,89,1965,99,1481,1563,
		809,1515,294,1474,1017,1325,150,817,1838,984,2025,1161,870,997,1478,1842,1665,1159,2010,97,1571,264,1755,1341,1140,1014,1569,428,119,1589,
		420,834,1841,1148,1019,1018,600,114,1479,426,299,298,1320,277,1928,84,2115,439,1865,1834,803,839,804,118,1505,1564,263,1497,1345,1949,2055,
		1595,1924,983,1857,1165,1859},
{1439,1649,895,1259,1458,1239,1079,1122,1419,359,1908,1249,659,1553,169,1195,2088,535,899,1110,1231,653,1138,1061,459,34,1053,539,1310,897,179,1828,
		1243,719,1132,1425,1363,1734,1055,1001,1914,1245,1007,1290,521,993,1915,171,1017,2089,883,997,2098,877,1003,1636,1415,1013,1646,1413,1009,
		1136,1433,839,1312,885,1183,2009,1069,1011,1648,525,999,1827,1257,829,410,151,271,2003,1063,1179,1316,353,1375,2095,1059,995,400,1237,1015,
		930,1241,1351,758,1071,1175,2090,1051,1353,1913,163,473,2099,1251,1173,1294,1067,91,1737,1427,1197,1731,1065,287,1546,177,1005,2097,511,631,
		2005,1075,297,1736,1073,1377,30,355,1359,1314,533,277,2096,1077,813,1126,537,1357,1912,347,1177,1300,161,93,1124,711,643,1907,887,1199,1909,
		1437,991,1998,1411,811,1457,1421,657,1735,1235,641,1637,167,109,1459,1057,463,1732,513,1379,1114,343,1355,1551,339,1019,1456,1423,823,1647,
		1255,283,1645,1435,1365,576,701,99,1130,1233,1181,1738,699,453,1547,349,1189,1555},
{125,218,33,284,6,148,43,101,230,9,272,54,18,7,29,124,177,30,184,127,35,52,5,255,153,160,76,79,115,231,173,28,129,19,267,77,236,121,149,4,105,139,
		150,53,266,224,109,100,17,244,103,279,65,117,64,254,41,256,174,55,141,185,78,31,88,242,113,57,219,89,175,157,102,21,243,137,81,206,151,67,
		145,161,45,194,163,189,207,278,91,169,126,165,187,172,42,181,11,195,260,1,208,47,34,248,23,237,73,186,25,226,261,10,136,93,225,37,196,162,
		249,238,112,69,90,13,49,273,138,107,92,212,131,66,97,268,285,85,119,114,68,200,143,128,133,201,280,16,213,155,61,40,176,27,220,104,59,80,179,
		15,258,152,221,166,232,130,8,191,83,116,123,281,56,3,222,233,190,32,167,71,269,246,39,282,239,275,20,95,154,99,75,270,94,234,106,287,202,257,
		111,87,178,252},
{475,1060,1359,1183,461,93,534,642,1072,402,1171,1092,508,1175,1371,1491,717,1195,340,999,463,281,453,894,462,172,222,811,372,2054,95,1191,1761,1437,
		655,1420,279,103,1361,1173,1254,1002,1432,762,91,1716,1694,815,111,1581,1375,400,823,641,639,1252,1362,42,654,273,455,451,1626,1228,831,1851,
		474,761,953,643,1118,1164,1001,1877,1182,1353,1169,511,1446,275,150,651,1671,1194,401,233,1363,1298,624,101,102,1697,993,449,1231,12,635,
		1230,291,2121,1374,593,41,938,1003,1344,821,1517,822,633,89,871,1986,1355,1515,471,2031,115,760,283,444,1181,398,813,282,1165,1074,995,151,
		732,1766,809,1875,767,1011,1941,835,700,1423,445,758,873,1014,1121,2076,65,112,1695,629,1881,837,1009,90,114,581,1313,625,218,880,523,1143,
		2057,1896,1145,870,1349,391,1017,1189,450,834,1301,413,1345,38,160,883,333,874,552,785,2055,269,751,1170,297,649,294,773,221,578,520,505,
		1243,295,153,1084,1806,510,989},
{63,1874,1329,833,1153,963,1784,69,293,433,783,2054,969,113,1213,243,1754,489,1013,493,803,1934,669,993,673,423,1484,249,653,1333,1143,2144,1029,
		1373,613,1323,2024,1149,83,1393,603,1574,609,273,323,1844,789,813,1043,2114,429,93,983,1514,849,453,263,1664,309,633,843,1964,1209,1193,1023,
		1604,129,473,303,1694,1389,832,123,209,1856,339,1163,929,1576,793,1353,443,29,1936,73,1173,623,1289,1831,973,1253,1343,569,1471,253,533,143,
		749,1486,1017,863,1109,1846,657,1921,389,1403,1197,1561,1049,683,837,1741,869,503,103,2101,629,1223,823,1651,329,1073,292,717,2011,1349,353,
		1012,1437,1841,809,1053,1419,112,1481,1229,873,1359,1496,1383,89,1676,999,483,989,1946,639,1203,509,1586,699,663,269,2036,99,1571,449,333,
		1059,472,1931,689,153,1179,1192,1661,149,1433,692,2021,1409,713,512,1751,1169,1413,652,159,2111,1454,693,1239,1372,805,2116,687,2126,2084,
		1233,265,1756,507,1766,1994,513,985,2026,1227,262},
{162,293,249,55,719,192,281,556,42,53,189,295,599,486,341,392,282,173,175,369,710,516,478,616,402,413,415,129,680,432,238,32,681,113,41,265,475,429,
		666,629,152,621,353,161,205,115,309,636,509,496,222,233,235,9,620,12,676,342,473,221,69,355,252,596,272,102,273,716,163,201,251,385,133,474,
		462,153,536,283,81,371,325,313,114,561,33,656,445,403,261,131,13,234,501,393,401,25,159,312,43,441,539,354,651,213,101,85,619,696,223,141,
		650,531,453,461,145,279,72,103,381,590,711,93,560,264,574,606,463,21,591,333,500,84,8,546,343,321,459,174,576,530,522,193,191,411,218,219,54,
		73,642,220,506,171,717,11,38,39,294,433,324,40,686,291,657,132,98,99,384,160,373,566,597,51,372,559,414,339,144,460,253,626,537,471,499,399,
		552,280,443,687,37,231,569,158,398,672,400,203,157,627,351,689},
{263,462,267,383,342,27,83,42,327,323,222,325,255,143,102,145,315,23,402,385,435,443,282,265,203,162,445,243,312,305,423,125,372,192,183,185,252,72,
		3,65,332,303,365,324,432,363,5,84,12,63,245,204,132,123,425,384,37,442,277,25,392,641,41,85,157,144,671,461,205,444,397,701,281,147,24,142,
		221,207,264,521,101,87,212,22,161,447,152,322,401,387,452,611,341,117,32,82,371,297,92,491,251,177,581,11,57,262,71,417,202,131,237,382,311,
		477,551,191,357,2,369,26,62,309,206,182,249,326,422,189,146,302,9,446,362,129,493,242,429,266,122,69,386,292,431,673,232,695,86,352,575,643,
		472,515,523,52,485,583,53,605,582,703,173,430,672,613,233,665,552,553,293,130,492,356,353,250,612,176,113,190,522,416,413,10,642,476,473,370,
		702,236,691,340,296,481,40,56,541,160,116,601,100,124},
{125,218,497,9,594,18,365,54,819,101,230,545,33,450,42,78,373,590,29,124,825,45,333,6,139,52,582,5,255,849,21,381,76,91,231,818,173,28,353,129,43,
		402,30,77,224,630,149,4,305,105,19,642,150,53,377,388,100,489,7,117,753,174,67,665,49,628,113,254,481,31,141,729,65,219,267,521,73,843,137,
		242,529,41,153,79,279,538,115,657,157,846,243,537,17,55,177,546,266,103,236,169,398,194,547,127,189,89,207,801,284,145,588,206,539,151,165,
		102,498,272,35,525,181,751,278,531,175,11,195,530,64,244,559,747,126,483,88,23,769,196,148,163,383,746,172,491,107,114,792,208,565,184,396,
		47,542,185,237,499,256,306,847,162,131,160,527,793,492,161,225,796,541,57,40,444,569,354,273,302,260,345,90,81,640,842,285,533,601,681,540,
		248,297,66,69,629,848,186,575,649,727,705,486,136}},
{{101,1015,2593,1173,2591,3492,2478,3365,2263,635,3354,2227,3399,2453,100,3736,2602,823,2431,3739,2270,3551,2403,2296,3313,2844,1579,2425,4008,4274,
		2604,3865,3223,2621,2745,2791,2451,2374,2959,3321,4039,3033,3094,3275,1608,821,295,2773,2239,335,2298,2223,3312,3125,2401,3011,2969,2443,
		2407,820,3649,2243,2476,2450,1355,3163,2664,2432,3747,4098,2646,3373,2784,98,3371,3043,3508,2413,2325,3955,2801,2565,2851,2271,103,2194,3139,
		4129,2914,1878,281,835,2650,3721,2843,2681,3522,2221,519,2283,2455,527,3493,3636,2273,2238,2420,2645,460,1755,3431,3534,3307,3843,3884,455,
		1968,278,3749,2235,3912,3852,4266,2975,2505,3467,2205,3769,1003,3617,2419,3582,1001,115,3533,1054,2812,2861,2463,3822,2454,3485,4176,2423,
		2825,2418,3579,2587,3679,275,879,1490,2275,1351,2826,2240,3403,2233,998,3942,3133,3155,2458,4185,2265,3287,2728,3203,2385,2611,3707,283,3501,
		3402,1518,815,3303,2771,3821,818,1375,2965,3545,2983,2690,3646,2415,2508,4197,1061,3758,2653,2767,3726,3794,2303,4086,3515,3496,2551},
{973,4000,3992,3644,3481,813,3162,2265,2403,3649,253,4090,4082,3734,3301,93,2982,2445,3739,2223,73,2960,2944,4011,993,3651,3331,61,3724,3659,2415,
		2433,793,3140,4101,273,3636,2409,3511,3634,3749,2235,2253,823,2945,2228,2262,4105,3996,1353,2458,1013,103,3125,2408,2442,4086,2990,453,2278,
		833,1003,4007,3132,3170,2438,961,3741,633,2428,283,4097,2952,4015,2276,241,2258,1173,2248,835,1560,4009,2231,2232,2244,1323,1015,1920,4099,
		2411,2412,3732,1143,295,1470,2975,2778,781,3726,423,265,2949,1935,3994,3147,3657,2246,603,85,3129,1830,4091,3155,991,2967,3748,783,985,3146,
		960,2951,271,3307,2456,963,805,2263,2418,242,3154,3487,91,3733,63,115,60,969,4001,3131,2274,3175,3643,243,1851,789,962,3124,2424,3658,2965,
		83,1941,69,782,3647,3747,2995,2426,1491,249,62,3737,2454,3145,1581,1476,3127,2254,811,822,792,2947,2434,3642,102,1566,2978,1485,2229,282,
		3722,247,1921,2763,1002,72,3158,1575,2417,252,2437,1865,4016,2590,1926,2257,1505,4106,2770,972},
{3317,992,2811,3497,1486,2335,2235,2263,2515,2415,2443,2233,2315,3321,2657,2631,3171,2371,2495,3501,2837,3602,2991,2957,3169,2613,2191,2989,2793,990,
		169,2983,3335,2831,163,2344,3053,3531,2967,3237,1439,3331,3147,3057,3051,3351,2184,1259,3511,3137,3163,3233,3155,3367,3612,1079,2681,3529,
		2164,3231,2975,3702,359,2861,3349,3692,659,2937,3381,3397,1195,3117,3547,3577,899,1649,3561,2364,653,1908,1061,2675,459,1553,1053,3581,539,
		1458,895,3587,179,1122,1239,719,34,1231,3515,2651,1055,2583,2633,1051,2763,2813,1065,2229,2803,1071,2409,2623,1059,2245,2271,1421,2425,2451,
		1437,2871,2627,2307,3413,2807,2639,3593,2949,3355,897,2405,1828,355,2691,2819,298,347,2325,3535,968,161,3037,3129,2903,1075,3217,2273,2413,
		151,2505,2453,2487,1069,1363,2088,1067,1001,1138,1243,1007,1132,521,993,1734,1419,1017,1110,171,997,1310,535,1003,1914,2341,1013,1915,883,
		1009,1290,1415,839,2089,1413,1183,2098,353,1011,1636,525,999,1136,1249,829,1646,887,271,2009,1425,1179,1827,1063,1375,1312},
{413,567,319,218,512,29,518,548,389,449,341,318,463,343,461,566,5,367,364,352,438,230,295,284,415,304,437,572,542,293,127,488,317,555,329,76,305,462,
		401,482,494,125,543,365,292,52,530,33,272,316,412,331,101,495,460,9,306,4,160,405,531,426,173,414,45,554,7,28,390,184,453,307,335,573,342,
		124,17,519,129,560,77,425,549,6,507,377,149,366,333,436,43,100,537,489,105,473,448,30,254,41,391,561,397,294,472,53,54,117,536,207,355,242,
		113,525,369,388,219,79,177,474,224,102,439,206,137,506,402,231,31,21,469,483,236,267,255,89,103,153,513,266,450,78,424,325,321,353,243,161,
		55,57,524,115,278,400,445,340,354,309,323,194,427,81,378,139,279,175,65,145,172,64,141,403,150,185,330,244,151,19,148,429,88,417,126,477,501,
		91,373,311,248,189,379,34,174,357,475,500,73,18,195,165},
{475,3131,3503,3595,3193,3341,894,2815,4216,3055,463,3072,2253,2558,2591,2785,3562,2948,3125,2971,1092,2163,3161,4130,3897,2724,2334,3956,3676,1195,
		2921,2853,3535,2983,3581,2484,2771,2168,2875,3235,4306,3128,1183,2892,3927,2845,4301,1171,1716,2343,3220,2544,3323,3687,655,2951,3073,1000,
		3415,103,2635,3521,2995,2382,3392,2433,3889,1017,2945,3361,2605,732,2981,3966,2814,3403,2514,3223,2745,1371,3285,3169,1375,3491,2673,3355,
		3013,1060,3043,3777,2323,2738,2320,2202,4084,3113,3572,2665,3041,1626,811,4056,1175,3502,2754,837,2865,2989,2472,474,3311,3175,3133,2499,
		1165,643,3401,2503,401,2183,2652,2574,3126,993,3523,3221,91,372,415,3728,4271,1191,2274,3049,624,2232,717,1194,2891,3383,1360,1359,445,2982,
		2454,2443,3343,3265,2260,297,2583,3103,2439,461,3807,451,1446,462,595,95,3921,3085,3101,111,3139,1344,1437,1374,3071,3123,3461,2682,823,1164,
		2695,2319,881,3445,2552,3994,1002,3657,2923,3065,295,2233,3837,1231,1986,3766,275,3987,831,3229,654,3551,1363,2953,625,279},
{2603,993,2469,929,2813,273,2499,209,2633,243,2259,389,2783,93,2229,1109,2253,2319,293,1289,2433,2439,1013,749,2613,2943,803,2697,2979,2793,3123,83,
		2637,2289,2403,2409,833,29,2223,3183,983,2877,2421,2853,813,2817,3159,3177,2673,1153,263,2997,1049,2993,963,2369,3237,113,3573,3003,4273,653,
		2769,3393,433,2517,1373,1349,3173,783,3029,3057,1193,3233,63,2969,3147,473,2693,453,269,2829,832,2277,2643,1043,2589,329,2397,2823,323,2649,
		1169,2487,3303,1333,2937,292,989,3483,613,2217,1012,89,3333,253,3117,1015,3053,112,3513,973,2457,295,2271,1856,2663,1173,629,2799,2815,2463,
		3039,533,809,2337,2283,3219,1253,569,2247,2493,1163,3189,509,2755,2313,443,3519,869,2967,2843,1353,449,2963,2427,3363,143,3339,1229,2635,
		3543,863,3129,689,2575,2873,633,149,3023,1496,2367,2583,623,3217,1409,992,2907,2763,1343,2977,3209,92,2349,2483,423,2909,1946,3203,2785,3501,
		1323,2399,1586,3027,1143,3489,2875,1676,2849,603,3309,2729,2036,2169,2961,812,2489,1403,2451},
{973,893,821,1396,1149,1392,1289,283,745,753,1295,1356,1165,1345,1063,1416,1077,1033,813,761,1013,163,1426,759,1422,1319,1265,972,1105,1375,1003,853,
		1386,1153,953,911,1112,1164,249,1409,1274,1183,1280,1032,1349,1070,1246,55,1092,1093,53,1310,162,1104,189,851,1216,958,1172,1073,312,1018,
		1010,295,1152,733,1376,1121,113,819,1135,1411,898,971,910,1326,1407,1330,1078,913,793,1346,1181,353,939,1381,1031,763,850,1138,1304,1437,
		1360,1379,252,943,293,1385,115,729,516,1421,1204,1082,1256,1439,880,928,1134,980,719,883,173,1415,475,789,192,1102,1089,1142,1226,1198,940,
		868,1194,1040,599,1250,1123,233,1151,369,1316,486,415,1061,1120,478,985,820,42,833,1286,309,1266,1162,355,432,1400,238,1045,1220,282,803,852,
		429,1306,1391,175,12,752,1139,1267,1355,760,402,413,1001,129,1276,235,666,1091,1430,914,1340,222,473,912,9,1325,731,782,636,941,474,1297,
		1370,342,743,1041,1296,881,1201,791,1212,696,114,1060,681},
{263,462,267,383,342,27,83,42,327,323,222,325,255,143,102,145,315,23,402,385,435,443,282,265,203,162,445,243,312,305,423,125,372,192,183,185,252,72,
		3,65,332,303,365,324,432,363,5,84,12,63,245,204,132,123,425,384,37,442,277,25,392,641,41,85,157,144,671,461,205,444,397,701,281,147,24,142,
		221,207,264,521,101,87,212,22,161,447,152,322,401,387,452,611,341,117,32,82,371,297,92,491,251,177,581,11,57,262,71,417,202,131,237,382,311,
		477,551,191,357,2,369,26,62,309,206,182,249,326,422,189,146,302,9,446,362,129,493,242,429,266,122,69,386,292,431,673,232,695,86,352,575,643,
		472,515,523,52,485,583,53,605,582,703,173,430,672,613,233,665,552,553,293,130,492,356,353,250,612,176,113,190,522,416,413,10,642,476,473,370,
		702,236,691,340,296,481,40,56,541,160,116,601,100,124},
{989,1143,497,895,218,1088,825,125,1094,1124,1025,917,894,1241,1039,919,1037,1142,489,893,1014,594,1070,77,230,1385,991,127,1058,928,977,1148,965,
		1119,1355,943,28,940,642,103,1024,21,76,869,871,1118,545,941,1131,882,101,849,1193,52,892,1038,1405,909,988,29,1403,9,1125,729,868,54,1425,
		1036,1071,266,5,990,1347,124,33,402,1107,881,1387,1137,911,918,1013,6,1345,1136,450,7,1002,4,45,1095,1419,1083,1641,173,942,1393,129,966,753,
		1048,43,100,53,1113,1179,1569,1065,1106,1401,31,105,1064,1050,30,1218,776,1177,148,883,353,1112,149,207,1201,1101,141,1149,1410,1082,115,538,
		1369,278,880,931,305,254,1395,905,967,1130,870,1433,1362,139,845,55,1026,78,243,242,1353,17,978,1015,79,1170,964,231,793,1429,102,1417,117,
		255,1411,1012,19,153,1089,1665,172,1377,769,1059,175,219,1213,1702,194,481,1001,930,177,1153,1545,150,916,137,768,1181,1623,151,1233,206,529,
		113,1203}}};
uint32_t usrparam=0;
uint32_t usrparam_mask=0;
volatile sig_atomic_t signalflag=0;
#if (MEMCHECK == 1) || (MEMCHECK == 2)
uint64_t allcdmemory=0;
#endif
pid_t *pids=NULL;
uint64_t memorylimit=0xFFFFFFFFFFFFFFFF;
uint64_t initavmemory=0;
void *mmapptr=NULL;
char *logpath=NULL;
int32_t *fvidx=NULL;
uint32_t usrseed=0;
uint32_t seeds[PBMTYPES]={1073741786,1073741770,1073741810,1073741770,1073741836,1073741770,1073741890,1073741756,1073741770}; /* With strategy number */
//uint32_t seeds[PBMTYPES]={1073741876,1073741836,1073741756,1073741890,1073741756,1073741756,1073741890,1073741800,1073741890}; /* Without strategy number */
#if defined(DEBUGCODE) && defined(VERBOSE)
int32_t verbfrom=0;
int32_t verbto=0x7fffffff;
#endif
#endif

/*----------------------------------------------------------------
 *
 *  Global variables shared with other modules that are not initialized here.
 *  WARNING: Don't use "static" as then a different variable will be created in
 *  each module with the same name.
 *
 *--------------------------------------------------------------*/
VARTYPE kbase kbset; /* Working KB. */
VARTYPE kbase mainkb; /* Main KB with axioms (theory) being analyzed. */
VARTYPE int num_cores;  /* Number of physical cores available. */
VARTYPE int active_cores;  /* Number of enabled cores */
VARTYPE int32_t procnb; /* Internal process number of current process. It is in the range 0 to active_cores-2 */
                        /* for child processes and equals active_cores-1 for parent process. */
VARTYPE int proc_affinity; /* Core affinity option. */
VARTYPE uint64_t numskolem; /* Number of skolem functions defined in main KB so far. */
VARTYPE int32_t skolemid; /* Number to identify unique skolem prefix. */
VARTYPE char skfprefix[15]; /* Prefix for unique skolem function names. */
VARTYPE uint64_t numpdefs; /* Number of predicate definitions in main KB so far. */
VARTYPE int32_t pdefid; /* Number to identify unique predicate definition prefix. */
VARTYPE char prdprefix[15]; /* Prefix for unique predicate definition names. */
VARTYPE int32_t splid; /* Number to identify unique P-predicate prefix. */
VARTYPE char splprefix[15]; /* Prefix for unique P-predicates. */
VARTYPE uint32_t numsymbols; /* Number of symbols for symbol weight assignment. */
#if PRECEDENCE != 0
VARTYPE uint32_t maxusecount; /* Maximum value of usecount field in hashchain symbol elements. */
#endif
VARTYPE hashchain *hashkeys[HASHVALUES]; /* Hash keys indexed by hash values of symbol names. */
VARTYPE hashchcmp *hashsplit[HASHVALUES]; /*Hash keys indexed by hash values of clause variants. */
VARTYPE uint32_t hashpos[MAXPOSITIONS]; /* Hash values for item positions. */
VARTYPE hashchain equkey; /* Key corresponding to equality literal. */
VARTYPE int32_t binpsize; /* Effective size of binprefix structure. */
VARTYPE int32_t satclsize; /* Effective size of satclause structure. */
#ifdef SEMANTICTAUTOLOGY
VARTYPE int32_t ttlgclsize; /* Effective size of ttlgclause structure. */
#endif
VARTYPE int32_t mxtrfsize; /* Maximum formula size in bytes for formulas added to discrimination trees. */
VARTYPE mpctl *procctl; /* Pointer to multi-processing control data to be placed in mmap area. */
VARTYPE double cpu_time; /* Process CPU time used in seconds. */
VARTYPE double *secprc_cpu_time; /* Pointer to total child processes CPU time used in seconds to be placed in mmap area. */
VARTYPE sem_t *procsem; /* Pointer to set of process synchronization semaphores to be placed in mmap area. */
VARTYPE stats glblstats; /* GLobal statistics. */
VARTYPE options glblopts; /* Global options. */
VARTYPE stats *pbmstats; /* Pointer to problem global statistics to be placed in mmap area. */
VARTYPE int32_t pbmflags; /* Flags for current problem. See definitions above. */
VARTYPE int32_t pbmtype; /* Problem type number. */
VARTYPE int32_t pgmmode; /* Problem mode: INTERACTIVE or NONINTERACTIVE. */
VARTYPE int32_t slctordrsize; /* Number of elements in slctordr set. */
VARTYPE int32_t selectqueues; /* Effective number of clause selection queues. There are selectqueues weight clause */
                              /* selection queues and the same number of age selection queues. */
VARTYPE uint32_t *strategies; /* Pointer to strategies selected for problem being solved placed in mmap area. */
VARTYPE int32_t *nextstrat; /* Pointer to index to next free strategy in the strategies[] set. */
VARTYPE int32_t nbstrats; /* Total number of strategies in the strategies[] set. */
VARTYPE int32_t numlrnrecs; /* Number of record elements in set pointed by learnrecord. */
VARTYPE uint64_t *maxprcmemory; /* Pointer to process memory limit in mmap region to be used for saturation processes. */
VARTYPE uint64_t maxmemory; /* Memory limit to be used for saturation processes. This equals to *maxprcmemory. */
VARTYPE uint64_t mmapsize; /* Size of mmap shared memory. */
VARTYPE uint32_t stratflgs; /* Active strategy flags. */
VARTYPE int32_t lkhdcomplete; /* Completeness indicator of lookahead literal selections. See SelectType56Cpl() function */
                              /* global comments for details. */
VARTYPE cmprefix *frstglselect[2*SELECTQUEUES]; /* Pointers to first formula in queue for each clause selection queue, */
                                    /* each pointer is NULL if no formulas of that weight added yet. */
                                    /* Only passive clauses are linked to clause selection queues. */
VARTYPE cmprefix *lstglselect[2*SELECTQUEUES]; /* Pointers to last formula in queue for each clause selection queue, */
                                    /* each pointer is NULL if no formulas of that weight added yet. */
                                    /* Only passive clauses are linked to clause selection queues. */
VARTYPE cmprefix *frstselect[2*SELECTQUEUES][MAXWEIGHT]; /* Pointers to first formula in queue for each clause selection */
                                    /* queue and for each weight/age, each pointer is NULL if no formulas of that weight/age */
                                    /* added yet. Only passive clauses are linked to clause selection queues. */
VARTYPE cmprefix *lstselect[2*SELECTQUEUES][MAXWEIGHT]; /* Pointers to last formula in queue for each clause selection */
                                    /* queue and for each weight/age, each pointer is NULL if no formulas of that weight/age */
                                    /* added yet. Only passive clauses are linked to clause selection queues. */
VARTYPE hashchain *minterm,*mintermbk; /* Minimal term in term ordering and its shuffling recovery backup value. */
#ifdef DEBUGMEMORY
VARTYPE void *memptrs[DBGMEMELEMS]; /* Pointers to memory allocations that must be tracked when DEBUGMEMORY is defined. */
                            /* The actual size of this set must be adjusted for each case. */
VARTYPE int32_t memtrack[DBGMEMELEMS]; /* Status of each entry in memptrs[]. The actual size of this set must be equal to the */
                               /* size of memptrs[]. The content is: */
                               /* 0 -> Entry not used. */
                               /* 1 -> Entry for memory allocated but not yet freed. */
                               /* 2 -> Entry for memory allocated and freed. */
VARTYPE int32_t ndemods;               /* Number of entries used in the above sets. */
VARTYPE int32_t auxcounter;            /* Auxiliary counter. */
#endif
#ifdef DEBUGCODE
VARTYPE int32_t dbgnb; /* For setting appropriate breakpoint in Otter and Discount functions. */
#endif
#ifdef BENCHMARKING
VARTYPE double *bresults[3]; /* Pointers to benchmarking results to be placed in mmap area. */
VARTYPE int32_t *bncounter; /* Pointer to benchmarking counter to be placed in mmap area. */
VARTYPE char *benchfname; /* Benchmarking output file name. */
#endif

/*----------------------------------------------------------------
 *
 *  Platform independent functions common to all platform independent modules.
 *
 *--------------------------------------------------------------*/
/* Drodi module functions. */
int main(int argc,char *argv[]); /* Entry point. */
int32_t Initialize_pgm(void); /* Initialize program */
void Clean4Exit(void); /* Performs the needed cleaning before exit */
int32_t CheckAnswer(char *,char *); /* Confirm answer with user */
int32_t Create_KB(kbase *); /* Create a new KB */
int32_t Initialize_KB(kbase *); /* Initialize a KB */
void Delete_KB(kbase *); /* Delete a KB */
void Free_Formula(kbase *); /* Free formula and tree memory and initialize queue pointers */
void FreeInfFormulas(void); /* Free inferred formulas in main KB */
void Initialize_hash(void); /* Initialize symbol hash tables for names in input formulas */
void Initialize_hashcmp(void); /* Initialize symbol hash tables for component names */
int32_t Tell(char *,int32_t,int32_t); /* Add formula to main KB */
int32_t InputSentence(char **,int32_t *); /* Input formula lines */
int32_t CheckSyntax(char **,int32_t *,char **,nameschain *,int32_t); /* Check sentence syntax */
char *SkipBlanks(char *); /* Skips space characters */
int32_t ParseSentenceStart(char *,char **,nameschain *,int32_t); /* Get formula start type */
int32_t ParseVariableList(char *string,char **,nameschain *); /* Variable list parser */
int32_t AddNametoChain(char *,nameschain *,int32_t,int32_t); /* Add name to text formulas names chain */
int32_t ParseAtomic(char *,char **,nameschain *,int32_t); /* Atomic formula parser */
int32_t ParseName(char *,char **); /* Parses a variable, constant, function or predicate name */
int32_t ParseTerm(char *,char **,char **,nameschain *,int32_t *,int32_t); /* Parse term */
int32_t CheckDeclared(char *,nameschain *,int32_t,int32_t *,int32_t *); /* Check if name is locally declared */
int32_t Ask(char *); /* Include a negated conjecture in main KB */
void AddTxtFormula2KB(cmprefix *,int32_t,kbase *,int32_t); /* Add text formula to a KB */
void ResetStats(stats *); /* Reset statistics structure */
void AddInferenceStats(stats *,stats *); /* Add inference statistics */
int32_t TPTPQuantifiers(char **,int32_t *); /* Set scope of quantifiers */
float FormatMemory(float,char **); /* Format memory in KB, MB and GB */
char *JumpOverQuotes(char *); /* Jump over a quoted block */
void CatchSignal1(int); /* Catch program exception signals. */
#if (MEMCHECK == 1) || (MEMCHECK == 2)
char *strdupmn(char *); /* Memory tracking version of strdup() function */
void *mallocmn(size_t); /* Allocate and track total memory */
void *reallocmn(void *,size_t); /* Reallocate and track total memory */
void freemn(void *); /* Free and track total memory */
#endif
uint64_t GetAvailableMemory(void); /* Get available memory */
int32_t Init_mmap(void); /* Initialize mmap common memory area */
#ifdef DEBUGMEMORY
void myfree(void *,char *,int); /* Performs memory debug tasks and calls free() function */
#endif
#ifdef MEMORYTUNNING
int32_t SymMalloc(uint64_t); /* Allocate additional memory for MAXMEMORY tuning */
#endif

/* clausify module functions. */
cmprefix *PreprocessFormula(cmprefix *,char **,int32_t *,int32_t); /* Text formula pre-processing */
int32_t PutBrackets(lkedfitem **ptopitem); /* Insert parenthesis according to precedence */
int32_t SafeAppend(char **,int32_t *,char *,int32_t); /* Append a string safely */
cmprefix *Convert2PreNNF(lkedfitem **,cmprefix *,char **,int32_t *); /* Perform pre-nnf conversion */
cmprefix *Convert2NNF(lkedfitem **,cmprefix *,char **,int32_t *); /* Perform nnf conversion */
void DropImplication(lkedfitem *); /* Translate an implication */
void DropRevImplication(lkedfitem *); /* Translate a reverse implication */
lkedfitem *DropEquivalence(lkedfitem *,lkedfitem **); /* Translate an equivalence */
lkedfitem *DropXOR(lkedfitem *,lkedfitem **); /* Translate a XOR connective */
void DropNAND(lkedfitem *); /* Translate a NAND */
void DropNOR(lkedfitem *); /* Translate a NOR */
int32_t IndentNegations(lkedfitem *); /* Move negations inwards */
void DropQuantifiers(lkedfitem **); /* Delete quantifiers in formula */
void RemoveLinkedListChain(lkedfitem **,lkedfitem *,lkedfitem *); /* Remove a chain of linked list items */
int32_t Convert2CNF(char *,char **,char **,int32_t *); /* Convert sentence to CNF */
int32_t CnfOrCnf(char **,int32_t *,char *); /* Merge two CNF's separated by an OR operator */
lkedfitem *Text2LinkedList(char *); /* Convert formula text to linked list format */
void FreeLkedLstFormula(lkedfitem *); /* Free formula storage in linked list format */
void RemoveRedundantBrackets(lkedfitem **); /* Remove redundant brackets */
lkedfitem *RemoveAdjBrackets(lkedfitem *,lkedfitem **); /* Remove set of adjacent redundant brackets */
void SetPolarity(lkedfitem *); /* Set polarity of linked list items */
void AddItem2LkdLstF(lkedfitem **,lkedfitem **,lkedfitem **); /* Add item to formula in linked list format */
int32_t AddOpenBracket(lkedfitem **,lkedfitem **,lkedfitem **); /* Add open bracket to formula in linked list format */
int32_t AddQuantifier(char *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain **); /* Add quantifier to formula in linked list format */
int32_t AddName(char *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add name to formula in linked list format */
int32_t AddPredOrFunc(hashchain *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add a predicate or function in linked list format */
int32_t AddEquality(char *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add an (in)equality in linked list format */
int32_t AddEquRootTerm(char *,char **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add an (in)equality root term in linked list format */
int32_t DuplicateBlock(lkedfitem *,lkedfitem *,lkedfitem **,lkedfitem **); /* Duplicate a block of linked list items */
void InsertBlock(lkedfitem *block,lkedfitem *,lkedfitem *); /* Insert a block of linked list items */
lkedfitem *JumpOverBlockOrTerm(lkedfitem *,int32_t); /* Jump over a block or term */
cmprefix *MiniscopeFormula(lkedfitem **,cmprefix *,char **,int32_t *); /* Miniscope a formula */
int32_t RemoveUnneededQfiers(lkedfitem **); /* Remove redundant quantifiers */
int32_t MiniscopeQuantifier(lkedfitem *,lkedfitem **,struct timeval *); /* Miniscope a quantifier */
int32_t MiniscopeSegment(lkedfitem *,lkedfitem *,lkedfitem **,struct timeval *); /* Miniscope a formula segment */
int32_t MiniscopeBlock(lkedfitem *,lkedfitem *,lkedfitem *,lkedfitem **,struct timeval *); /* Miniscope a formula block */
int32_t IsVarInBlock(lkedfitem *,lkedfitem *); /* Check if a variable exists in a block */
void RelinkVars(lkedfitem *,lkedfitem *,lkedfitem *); /* Rebuild links of variables in a block */
lkedfitem *DuplicateQuantifier(lkedfitem *,lkedfitem **,lkedfitem **,nameschain *); /* Duplicate and insert a quantifier */
cmprefix *Skolemize(lkedfitem **,cmprefix *,char **,int32_t *); /* Skolemize a formula */
cmprefix *TFSimplify(lkedfitem **,cmprefix *,char **,int32_t *); /* Simplify $true and $false predicates */
void RenameVariables(lkedfitem *); /* Rename variables in a formula */
cmprefix * AddLkdLstFormula2KB(lkedfitem *,cmprefix *,cmprefix *,int32_t,char **,
		int32_t *); /* Add linked list formula to main KB in text form */
cmprefix *FormulaRenaming(lkedfitem **,cmprefix *,char **,int32_t *); /* Formula renaming */
void GetClauseNumberBounds(lkedfitem *,int32_t *,int32_t *); /* Get lower bounds of number of clauses in a segment */
void CalcSegmentBounds(int32_t *,int32_t *,int32_t,int32_t,int32_t,int32_t,
		int32_t); /* Calculate lower bounds of number of clauses in a segment */
int32_t SegmentRenaming(lkedfitem **,lkedfitem *,int32_t,int32_t,cmprefix **,
		cmprefix *,nameschain *); /* Rename blocks within a segment */
void GetaABCoefs(int32_t,int32_t,lkedfitem *,lkedfitem *,int32_t,int32_t *,
		int32_t *); /* Get "a" and "b" coefficients of a block */
int32_t RenameBlock(lkedfitem **,lkedfitem *,cmprefix **,cmprefix *,nameschain *); /* Rename a block in a formula */

/* compile module functions. */
int32_t CompileCNF(char *formula,cmprefix *); /* Compile a CNF text formula */
int32_t CompileClause(char *,cmprefix **,cmprefix *); /* Compile a clause */
void CompileLiteral(char *,int32_t *,cmprefix *,int32_t *); /* Literal compilation */
void CompileTerm(char *,int32_t *,cmprefix *,int32_t *); /* Term compilation */
void CompileSymbol(char *,int32_t *,cmprefix *,int32_t *); /* Symbol compilation */
int32_t BytesRequired(char *); /* Compute the number of bytes needed for a compiled clause */
int32_t AddBinClause2KB(cmprefix *,kbase *,int32_t); /* Add binary clause to a KB. */
int32_t AddUEQClause2KB(cmprefix *,int32_t); /* Add binary clause to a KB for unfailing completion */
int32_t RenumberVars(binprefix *); /* Renumber variables in a binary formula */
int32_t RenumberVars2(binprefix *); /* Renumber variables in critical pairs */
int32_t SafeAppendF(char **,int32_t *,char *,int32_t *); /* Append a string fast and safely */

/* indexing module functions. */
int32_t AddGlobalSymbols(nameschain *); /* Add locally defined symbols to global symbol list */
hashchain *CreateGlobalSymbol(char *,int32_t,int32_t); /* Create global symbol entry */
#ifdef MURMUR3
uint32_t Murmur3(char *); /* Murmur3 hash function */
#else
uint32_t KRMultHash(char *); /* Kernighan and Ritchie hash function */
#endif
hashchain *CheckGlblDeclared(char *); /* Check if a name is globally declared */
int32_t EditDiscrTrees(cmprefix *,int32_t); /* Edit discrimination trees for clause */
int32_t EditUEQTrees(cmprefix *,int32_t); /* Edit unfailing completion discrimination trees for clause */
#ifdef SEMANTICTAUTOLOGY
int32_t EditTtlgDscTrees(ttlgclause *,ttlgdstrleaf **,
		int32_t); /* Edit semantic tautology detection discrimination trees */
#endif
int32_t DsTermIndexing(uint8_t *,void **); /* Add item to a discrimination tree */
#ifdef SEMANTICTAUTOLOGY
int32_t DsTtlgIndexing(ttlgclause *,uint8_t *,void **); /* Add item to a semantic tautology discrimination tree */
#endif
void DeleteBranch(uint8_t *,void **); /* Delete item in a discrimination tree */
#ifdef SEMANTICTAUTOLOGY
void DeleteTtlgBranch(uint8_t *,void **); /* Query a semantic tautology discrimination tree */
#endif
void FreeDiscrBranch(void *,int32_t); /* Free discrimination tree */
#ifdef SEMANTICTAUTOLOGY
void FreeTtlgDiscrBranch(void *,int32_t); /* Free semantic tautology discrimination tree */
#endif
int32_t QueryDscTree(uint8_t *,uint8_t **,void *,int32_t,int32_t); /* Query a discrimination tree */
int32_t QueryDscTree2(uint8_t *,uint8_t **,void *,dstrbktrk *,
		int32_t); /* Query a discrimination tree by recursive function */
#ifdef SEMANTICTAUTOLOGY
int32_t QueryTtlgDscTree(uint8_t *,uint8_t **,uint8_t **,void **); /* Query a semantic tautology discrimination tree */
#endif
void LinkSelectQueues(cmprefix *); /* Link binary clause to clause selection queues */
void UnlinkSelectQueues(cmprefix *); /* Unlink binary clause from clause selection queues */
int32_t Unlink(cmprefix *,kbase *); /* Unlink binary clause from KB */
void UnlinkTxtFormula(cmprefix *,kbase *); /* Unlink text formula from KB */
int32_t UnlinkUEQ(cmprefix *); /* Unlink binary clause from KB for unfailing completion */
void FreeTreeMemory(void); /* Free KB tree memory */
#ifdef SEMANTICTAUTOLOGY
void FreeTtlgTreeMemory(ttlgdstrleaf **,int32_t); /* Free semantic tautologies tree memory */
#endif
uint32_t HashClause(uint8_t *); /* Calculate 32 bits hash value for a clause key */
#ifdef DEBUGTREE
void CheckDscTrees(void); /* Discrimination trees integrity check */
#else
#ifdef DEBUGTREE2
void CheckDscTQueries(void); /* Discrimination trees queries check */
#endif
#endif

/* infere module functions. */
uint8_t *NextItem(uint8_t *,int32_t); /* Jump to next item in clause */
int32_t IsEqual(uint8_t *,uint8_t *,int32_t); /* Check if two items are identical */
int32_t IsEqual2(uint8_t *,uint8_t *); /* Check if two items in different clauses are identical */
int32_t Unify(uint8_t *,uint8_t *,subst *,int32_t); /* Perform literal or term unification */
int32_t UnifyVar(uint8_t *,uint8_t *,subst *,int32_t); /* Unify a variable with a term */
int32_t SymEquUnify(uint8_t *,uint8_t *,subst *,int32_t); /* Perform symmetric equality unification */
int32_t Generalize(uint8_t *,uint8_t *,subst *,int32_t); /* Perform literal, term or list generalization */
int32_t GeneralizeVar(uint8_t *,uint8_t *,subst *); /* Generalize a variable with a term */
int32_t SymEquGeneralize(uint8_t *,uint8_t *,subst *,int32_t); /* Perform symmetric equality generalization */
int32_t IsCyclicSubst(subst *); /* Cyclic substitution check */
int32_t CheckClosedLoop(uint8_t *,uint8_t *,int16_t *,int32_t *,int16_t *,
		int32_t *); /* Check if a variable is in a closed loop */
uint32_t CheckVarInTerm(uint8_t *,uint8_t *,subst *,
		int32_t); /* Check if a variable occurs in a term with substitution */
uint32_t CheckVarInTerm2(uint8_t *,uint8_t *); /* Check if a variable occurs in a term */
int32_t CopyClause(cmprefix *,int32_t *,cmprefix *); /* Copy a clause */
void DeleteLiteral(binprefix *,uint8_t *); /* Deletes a literal in a clause */
uint8_t *GetLiteralPtr(cmprefix *,int32_t); /* Gets a pointer to literal given its order number */
int32_t GetLiteralNbr(binprefix *,uint8_t *); /* Gets a literal number given a literal pointer */
int32_t ApplySubst2Clause(cmprefix *,int32_t *,subst *,int32_t); /* Apply substitution to a clause */
void GetSubstTermData(subst *,uint8_t **,int32_t); /* Initialize set of pointers to terms in a substitution */
int32_t ApplySubst2Item(uint8_t **,binprefix **,int32_t *,uint8_t *,
		int32_t,uint8_t **); /* Apply substitution to an item */
void UpdateParams(binprefix *); /* Update binary clause parameters */
void UpdateParams2(uint8_t *); /* Update binary formula parameters */
uint8_t *UpdateSymbolParms(binprefix *,uint8_t *,int32_t); /* Update binprefix and symbol parameters in an item */
uint8_t *UpdateSymbolParms2(uint8_t *,uint8_t *); /* Update symbol parameters in an item */
void Infere(cmprefix *); /* Perform saturation inferences */
void Resolution(cmprefix *); /* Perform resolution inference */
int32_t BuildResolvent(cmprefix *,cmprefix *,uint8_t *,uint8_t *,
		cmprefix **,subst *,int32_t); /* Build clause inferred by resolution */
void PM2Clause(cmprefix *); /* Perform Paramodulation to a clause */
void PMFromClause(cmprefix *); /* Perform Paramodulation from a clause */
int32_t BuildPMClause(cmprefix *,cmprefix *,uint8_t *,uint8_t *,uint8_t *,uint8_t *,
		cmprefix **,subst *); /* Build clause inferred by paramodulation */
void EqualityResolution(cmprefix *); /* Perform equality resolution inference */
void EqualityFactoring(cmprefix *); /* Perform equality factoring inference */
void Factor(cmprefix *); /* Infer factors of a clause */

/* simplify module functions. */
int32_t IsTautology(cmprefix *); /* Check if clause is a tautology. */
int32_t RemoveDuplicates(cmprefix *,cmprefix **,int32_t); /* Remove literal duplicates from clause */
int32_t TrivEqResolution(cmprefix *,cmprefix **); /* Trivial equality resolution inference */
int32_t DestrEqResolution(cmprefix *,cmprefix **); /* Destructive equality resolution inference */
int32_t FwSubsuming(cmprefix *,int32_t); /* Check if clause is subsumed by a KB */
int32_t BckSubsuming(cmprefix *,int32_t); /* Perform backward subsumption inference */
int32_t IsSubsumedByClause(uint8_t *,int32_t,binprefix *,subst *,
		uint8_t **); /* Check if clause is subsumed by another clause */
int32_t LiteralSubsuming(uint8_t *,uint8_t *,int32_t,subsumdata *,subst *,
		struct timeval *,uint8_t **); /* Check literal subsuming */
int32_t FwdSimplify(cmprefix **,int32_t); /* Forward and backward simplifications */
int32_t FirstSimplify(cmprefix **,int32_t); /* First clause simplification */
int32_t SecondSimplify(cmprefix **,int32_t); /* Forward clause simplification */
int32_t BckSimplify(cmprefix *,int32_t); /* Backward simplification */
int32_t ResolFactor(binprefix *,uint8_t *,binprefix **,uint8_t **,
		subst *,int32_t); /* Factoring for resolution inferences */
int32_t FwdDemodulation(cmprefix **,int32_t); /* Perform forward demodulation */
int32_t BckDemodulation(cmprefix *,int32_t); /* Perform backward demodulation */
int32_t Demodulate(cmprefix *,uint8_t *,binprefix *,uint8_t *,
		cmprefix **,subst *); /* Demodulate a clause */
#ifdef REDUNDANCYCOMPLETE
int32_t RedundancyCheck(uint8_t *,binprefix *,binprefix *); /* Perform demodulation redundancy check */
#else
int32_t RedundancyCheck(uint8_t *,uint8_t *,binprefix *,binprefix *); /* Perform demodulation redundancy check */
#endif
int32_t FwSubsResltn(cmprefix *,cmprefix **,int32_t); /* Perform forward subsumption resolution */
int32_t BkSubsResltn(cmprefix *,int32_t); /* Perform backward subsumption resolution */
int32_t SubsResltn(cmprefix *,binprefix *,uint8_t *,cmprefix **); /* Perform subsumption resolution inference */
#ifdef UNITCLRESOLUTION
int32_t UCResolution(cmprefix *); /* Perform unit clause resolution inference */
#endif

/* ueqlib module functions. */
int32_t Normalize(cmprefix **,int32_t); /* Normalize an equation */
int32_t FwdNormalize(cmprefix **,int32_t); /* Forward simplification for UEQOTT and UEQDSC algorithms */
int32_t IsEqTautology(cmprefix *); /* Check if an (in)equality is an equational tautology */
int32_t IsCPConnected(cmprefix *); /* Check if a critical pair is connected */
int32_t IsConnected(uint8_t *,uint8_t *,uint8_t *,coeff *,
		int32_t,int32_t); /* Check if two terms are connected */
int32_t IsGrJoinable(uint8_t *,int32_t,int32_t,int32_t,int32_t); /* Check if an equality is ground joinable */
int32_t GJRewrite(int32_t *,int32_t *,int32_t *,uint8_t *,uint8_t *,
		int32_t,int32_t); /* Check if two terms are ground joinable */
int32_t FwUEQSubsum(cmprefix *); /* Check if an equality is subsumed */
int32_t BkUEQSubsum(cmprefix *); /* Backward sumbumption simplifications */
int32_t GenerateCP(cmprefix *); /* Generate critical pairs */
int32_t GenerateCPTo(cmprefix *); /* Generate critical pairs to clause */
int32_t GenerateCPFrom(cmprefix *); /* Generate critical pairs from clause */
int32_t MakeTrivialProof(cmprefix *); /* Generate a trivial proof for UEQ problem */
int32_t FwUEQSimplify(cmprefix **); /* Perform forward simplifications of UNPROC clause */
int32_t BckUEQSimplify(cmprefix *); /* Backward simplify active clauses */
int32_t UEQEqResolution(cmprefix **,int32_t); /* Equality resolution inference for unfailing completion */
int32_t UEQResolution(cmprefix **); /* Resolution goal simplification for unfailing completion */
uint8_t *ApplySubst2Term(uint8_t *,subst *,int32_t); /* Apply an instance substitution to a term */
void ShiftUEQVarNbr(binprefix *,int32_t); /* Shift unit equality variable numbers */
void UnshiftUEQVarNbr(binprefix *,int32_t); /* Unshift unit equality variable numbers */
int32_t NextPerm(int32_t *,int32_t *,int32_t *,int32_t); /* Generate next permutation */
int32_t SplitNAdd2Active(cmprefix **); /* Split unit equalities and add to ACTIVE */
uint8_t *ApplyFastSubst(uint8_t *,int32_t *,int32_t *); /* Apply a special fast substitution to a term */
int32_t CheckWeakRewrite(uint8_t *,uint8_t *,subst *,int32_t); /* Check if a weak rewrite is valid */

/* ordering module functions. */
int32_t CompareLiterals(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* Literals comparison */
int32_t CompareEqualities(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* Equality literals comparison */
int32_t ComparePredicates(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* Predicates comparison */
int32_t CompareItems(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* Terms and predicates comparison */
int32_t CompareItems2(uint8_t *,uint8_t *,coeff *,coeff *,int32_t,int32_t); /* Another items comparison */
int32_t CompareItems3(uint8_t *,uint8_t *,int32_t,int32_t); /* Just another items comparison */
int32_t CompareEquTerms(binprefix *,uint8_t *,uint8_t *,equpnode **,
		equpnode **,int32_t); /* Equality root terms comparison */
int32_t CompSubstEquTrms(binprefix *,uint8_t *,uint8_t *,equpnode **,equpnode **,subst *,
		int32_t,int32_t); /* Equality root terms comparison with substitution */
int32_t CompareSubstItems(uint8_t *,uint8_t *,int32_t,subst *,int32_t,
		int32_t); /* Terms comparison with substitution */
int32_t ConnectCompare(uint8_t *,uint8_t *,int32_t,int32_t); /* Terms comparison specific for connectedness check */
int32_t CheckGJOrdering1(int32_t *,int32_t *,uint8_t *,uint8_t *,
		int32_t); /* Check term ordering condition for ground joinability */
int32_t CheckGJOrdering2FE(int32_t *,int32_t *,int32_t *,uint8_t *,uint8_t *,
		int32_t); /* Front end for CheckGJOrdering2() function */
int32_t CheckGJOrdering2(int32_t *,int32_t *,int32_t *,uint8_t *,uint8_t *,
		int32_t); /* Check term ordering condition for ground joinability */
void Equalize(int32_t,int32_t,int32_t *,int32_t *,
		int32_t); /* Equalize two variables for ground joinability checking */
#ifdef SEMANTICTAUTOLOGY
int32_t TtlgCompare(uint8_t *,uint8_t *); /* Term comparison for semantic tautology detection */
#endif
int32_t SecondKBOCheck(uint8_t *,uint8_t *,int32_t); /* Perform second KBO check */
int32_t SecondKBOCheck2(binprefix *,uint8_t *,uint8_t *,equpnode *,
		equpnode *,int32_t); /* Perform second KBO check for polynomial trees */
int32_t SecondKBOCheck3(binprefix *,uint8_t *,uint8_t *,equpnode *,
		equpnode *,subst *,int32_t,int32_t); /* Perform second KBO check with substitution */
int32_t SecondKBOCheck4(uint8_t *,uint8_t *,int32_t,int32_t); /* Perform second KBO check */
		/* for UEQDSC and UEQOTT algorithms */
int32_t SecondKBOCheck5(uint8_t *,uint8_t *,int32_t,int32_t); /* Another second KBO check */
int32_t CompareLexic(binprefix *,uint8_t *,uint8_t *,equpnode *,equpnode *,subst *,int32_t,
		int32_t); /* Perform lexicographical comparison with substitution */
int32_t CompareItemWeight(uint8_t *,uint8_t *,coeff *,coeff *); /* Terms and predicates weight comparison */
int32_t CompareItemWeight2(coeffcpt *,coeffcpt *); /* Equality subterms weight comparison */
int32_t CompareItemWeight3(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* items weight comparison */
int32_t CompareItemWeight4(uint8_t *,uint8_t *,int32_t); /* Another items weight comparison */
int32_t CompItmWeightSubst(equpnode *,equpnode *,subst *,int32_t,
		int32_t);/* Equality subterms weight comparison with substitution. */
int32_t ApplySubst2Pol(lkedcoef **,subst *,int32_t,int32_t); /* Apply substitution to polynomial. */
int32_t ApplySbst2TwoItm(uint8_t *,uint8_t *,uint8_t **,uint8_t **,cmprefix **,
		subst *,int32_t,int32_t); /* Apply substitution to two terms */
int32_t CalcCoeffs(uint8_t *,coeff *,int32_t); /* Calculate weight polynomial coefficients */
int32_t CheckEqRootOrder(binprefix *,uint8_t *,uint8_t *,uint8_t *,subst *,int32_t,
		int32_t); /* Check ordering constraints in equality root terms */
int32_t CheckEqRootOrder2(uint8_t *,uint8_t *,subst *,int32_t); /* Check ordering constraints in equality root terms */
uint8_t *UpdateOffsets(binprefix *,uint8_t *); /* Update item offsets */
int32_t SetEquData(binprefix *,uint8_t *,uint8_t *,uint8_t *); /* Initialize equality data for ordering constraints. */
int32_t SetEquChildNode(binprefix *,uint8_t *,
		equpnode **); /* Initialize and set equality tree node for ordering constraints. */
int32_t SetEqTrmPol(binprefix *,uint8_t *,coeffcpt **); /* Set equality subterm compact polynomial. */
int32_t SetLftRhtPols(binprefix *,coeffcpt **,coeffcpt **); /* Convert compact polynomials to left and right format. */
void FreeEquData(binprefix *); /* Free equality ordering data storage allocated for a compiled clause */
void FreeEquNode(uint8_t *,equpnode *); /* Free equality ordering node and its data. */
void FreeEquNodeData(uint8_t *,equpnode *); /* Free data in an equality ordering node. */

/* selecting module functions. */
int32_t SelectItems(cmprefix *); /* Select items in a clause */
int32_t SelectLiterals1(cmprefix *,char *,int32_t,int32_t,int32_t); /* Several types of literal selection */
int32_t SelectLiterals2(cmprefix *,char *,int32_t); /* MXSINGLENEG literal selection */
int32_t SelectMaximal(cmprefix *,char *,int32_t,int32_t *,int32_t *,int32_t *); /* Auxiliary maximal literal selection */
int32_t SelectByWeight(cmprefix *,char *); /* Select literals by weight */
int32_t SelectByVars(cmprefix *,char *); /* Select literals by number of variables */
int32_t SelectByTop(cmprefix *,char *); /* Select literals by number of top level variables */
int32_t SelectByDVars(cmprefix *,char *); /* Select literals by number of distinct variables */
int32_t SelectByNoPosEq(cmprefix *,char *); /* Select literals by no positive equalities */
int32_t SelectByNegEq(cmprefix *,char *); /* Select negative equalities */
int32_t SelectByNegative(cmprefix *,char *); /* Select negative literals */
int32_t SelectByLMin(cmprefix *,char *); /* Select literals with minimal lookahead children */
int32_t SelectByLMax(cmprefix *,char *); /* Select literals with maximal lookahead children */
void SelectByLexic(cmprefix *,char *,int32_t); /* Lexical tie break literal selection */
int32_t CompareLexic4Select(uint8_t *,uint8_t *); /* Lexical comparison for SelectByLexic function */
int32_t SelectType1(cmprefix *,char *); /* Type 1 literal selection */
int32_t SelectType2(cmprefix *,char *); /* Type 2 literal selection */
int32_t SelectType3(cmprefix *,char *); /* Type 3 literal selection */
int32_t SelectType4(cmprefix *,char *); /* Type 4 literal selection */
int32_t SelectType5(cmprefix *,char *); /* Type 5 literal selection */
int32_t SelectType6(cmprefix *,char *); /* Type 6 literal selection */
int32_t SelectType14Cpl(int32_t (*selectfunc)(cmprefix *,char *),cmprefix *,
		char *,int32_t); /* Complete version of Type 1 to 4 literal selection */
int32_t SelectType56Cpl(int32_t (*selectfunc)(cmprefix *,char *),cmprefix *,
		char *,int32_t); /* Complete version of Type 5 and 6 literal selection */

/* multith module functions. */
void Saturate(void); /* Start saturation algorithms */
int32_t SetMTParams(int32_t); /* Set strategy specific KB parameters */
void InferenceProcess(void); /* Main inference function for the process */
void Otter(void); /* Otter saturation algorithm */
void Discount(void); /* Discount saturation algorithm */
void UEQOtter(void); /* Otter saturation algorithm for unfailing completion */
void UEQDscnt(void); /* Discount saturation algorithm for unfailing completion */
cmprefix *SelectClause(void); /* Select current clause from passive clauses */
int32_t Retained(cmprefix *); /* Check if clause is retained */
void ClauseWeight(cmprefix *); /* Compute clause weight */
int32_t LRSClean(int32_t); /* Delete unreachable passive clauses */
int32_t SetClSelectOrder(void); /* Set select order for clause selection queues */
int32_t Preprocess(int32_t); /* Perform problem pre-processing */
uint64_t ClauseRelevance(void); /* Perform relevance filtering */
int32_t InitSInEDist(void); /* Perform SInE processing */
void GoType(void); /* Print problem type flags */
void Clausify(void); /* Convert problem formulas to CNF and print them in TPTP format */
int32_t InitStrategies(void); /* Select strategies for current problem */
void GetPbmTypeNumber(void); /* Calculate problem type number */
uint32_t StratNb2StratFlags(int32_t); /* Convert a strategy number to strategy flags */
uint32_t LookAhead(cmprefix *); /* Estimate number of inferences of a given clause */

/* filemgmt module functions. */
char *Decompile(binprefix *); /* Translate a binary clause formula to text format */
char *Decompile2(uint8_t *); /* Translate a binary hashchcmp formula to text format */
char *DecompileSat(satclause *); /* Translate a SAT clause formula to text format */
char *DecompileAsserts(uint8_t *,int32_t); /* Translate an A-clause assertions to text format */
int32_t AddTermText(uint8_t *,char **,int32_t *,int32_t *); /* Translate a binary term to text format */
#ifdef DEBUGCODE
void BuildTxtSubstData(uint8_t *,uint8_t *,subst *,int32_t); /* Build substitution data in text form */
#endif
void PrintStats(int32_t); /* Display statistics */
void PrintStatData(stats *,int32_t); /* Display statistics data */
void PrintProcessSettings(void); /* Print process option settings */
int32_t BuildProof(void); /* Print theorem proof */
int32_t AddSatCl2Proof(void); /* Add SAT clauses to proof queue */
int32_t AddAclause2Proof(proofnode *); /* Add A-clause to proof queue */
void LinkProofNode(proofnode *,proofnode *); /* Link node to proof queue */
int32_t PrintProof(void); /* Print theorem proof */
int32_t PrintModel(void); /* Print model from a GOSAT run */
int32_t Import(char *,char *,int32_t,int32_t); /* Import a TPTP problem file */
int32_t ReadDirective(char **,int32_t *,FILE *); /* Read a directive from a TPTP file */
int32_t CheckBrackets(char **); /* Check brackets and parenthesis balance */
int32_t Cnf2Fof(char **); /* Convert a CNF formula to FOF */
int32_t FileExist(char *); /* Check if a file exist */
void PrintKBData(void); /* Print current data of a knowledge base */
int32_t PrintAClauses(void); /* Print A-clauses of a knowledge base */
int32_t PrintSingleAClause(cmprefix *); /* Print a single A-clause */
int32_t PrintCompSymbols(void); /* Print the component symbols of a knowledge base */
int32_t PrintSATClauses(void); /* Print SAT clauses of a knowledge base */
int32_t PrintSingleSATClause(satclause *); /* Print a single SAT clause */
int32_t LinkedList2Text(lkedfitem *,char **,int32_t *,int32_t); /* Convert formula in linked list format to text */
int32_t Equ2Text(lkedfitem **,char **,int32_t *,int32_t *); /* Convert (in)equality in linked list formula to text */
int32_t PredOrFunc2Text(lkedfitem **,char **,int32_t *,
		int32_t *); /* Convert predicate or function in linked list formula to text */
int32_t Term2Text(lkedfitem **,char **,int32_t *,int32_t *); /* Convert term in linked list formula to text */
int32_t Quantifier2Text(lkedfitem **,char **,int32_t *,int32_t *); /* Convert a quantifier in linked list formula to text */
int32_t PrintClausifiedFormulas(int32_t); /* Print main KB formulas in TPTP CNF format */
char *ParseFilename(char *,char **); /* Parse and build a filename */
void Help(char*); /* Display general help or a help topic */

/* satsolver module functions. */
int32_t SplitClause(cmprefix *,int32_t); /* Get maximal split of an A-clause */
int32_t AddAssertions(binprefix *,binprefix *,binprefix *); /* Add assertions from two clauses to a third clause */
int32_t GetLocksAsserts(binprefix *,binprefix *,binprefix *); /* Get simplification assertions and locking assertions */
void ChainAsserts(uint8_t *); /* Chain assertions to appropriate queue */
void ChainLocks(uint8_t *); /* Chain locking assertions to appropriate queue */
void UnchainAsserts(uint8_t *); /* Unchain assertions */
void UnchainLocks(uint8_t *); /* Unchain locking assertions */
int32_t GetSplitSymbol(uint8_t *,int32_t,int32_t,hashchcmp **,int32_t); /* Get a P-predicate symbol for a clause */
int32_t IsVariant(uint8_t *,int32_t,hashchcmp *); /* Check if clause is a variant of another clause */
int32_t LiteralVariant(uint8_t *,hashchcmp *,int8_t *,int32_t,subst *,struct timeval *); /* Literal variant check */
int32_t ItemVariant(uint8_t *,uint8_t *,subst *); /* Item variant check */
int32_t SatQ2SatSolver(int32_t); /* Send SAT queue to SAT solver */
int32_t SatSolver(void); /* SAT solver algorithm */
void ChainOrdComponent(hashchcmp *); /* Chain a hashchcmp element to the ordered sub-queue */
void UnchainOrdComponent(hashchcmp *); /* Remove a hashchcmp element from the ordered sub-queue */
void ChainSatClause(satclause *); /* Chain a SAT solver clause to ordered sub-queue */
void UnchainSatClause(satclause *); /* Remove a satclause element from the ordered sub-queue */
int32_t DoSymbolAssgnmnt(satnode *,satclause **); /* Perform symbol assignment tasks */
void UndoSymbolAssgnmnt(satnode *); /* Undo symbol assignment */
void GoSat(void); /* Call SAT solver without first order prover */
int32_t ApplyModelChanges(void); /* Apply changes of current interpretation */
int32_t RemoveLocks(hashchcmp *); /* Remove locks unlocked by model change */
int32_t DoLocks(hashchcmp *); /* Do new locks produced by model change */
int32_t AddCompClause(hashchcmp *); /* Add non existing component clauses */
int32_t BuildContrClause(cmprefix *); /* Add contradiction clause to SAT queue */
int32_t SatSimplify(satclause **); /* Check tautologies and remove duplicates in SAT clause */
int32_t SatSubsumption(satclause *); /* Perform by and to Sat clause subsumptions */
int32_t IsSatClSubsumed(satclause *,satclause *,uint8_t *); /* Check if a Sat clause subsumes another Sat clause */
int32_t Backtrack(satclause *,satnode **); /* Backtrack from a contradiction */
#ifdef VERIFYSATPROOF
void VerifySatProof(void); /* Verify SAT part of the proof */
#endif
#if VERIFYSATFINALSTATUS == 1
void VerifySatStatus(void); /* Verify SAT final status after a problem solving run */
#endif

/* learn module functions. */
int32_t ProcessLearnData(char *,char *); /* Build learn data from output learn file */
learnvector *DCDescentMethod(learnvector *,uint8_t *,int32_t,int32_t,float *,float *,
		float *,float *,int32_t *,int32_t *); /* Dual Coordinate Descent Method */
void GetAccuracy(learnvector *,uint8_t *,int32_t,int32_t,float *,float *,
		float *,int32_t *,int32_t *); /* Get learning accuracy data */
learnvector *GetClauseVector(cmprefix *); /* Build a clause vector */
int32_t VerticalFeatures(cmprefix *); /* Compute vertical features of a clause vector */
int32_t  HorizontalFeatures(cmprefix *); /* Compute horizontal features of a clause vector */
int32_t  StaticFeatures(cmprefix *clause); /* Compute static features of a clause vector */
void LearnFromProblem(); /* Extract learning information from solved problem */
int32_t ReadLearnRecords(void); /* Read learn records generated by PROCLEARN command */

/* parsing module functions. */
void CommandProc(void); /* Command processor */
void ParseArgs(int,char **); /* Parse program arguments */
int32_t ChkConvertOption(char *); /* Check and convert option to interactive command */
int32_t ParseCommand(char *,char **); /* Command parser */
#endif /* GLOBAL_H_ */

/* Shuffling and randomizing functions. */
int32_t ShuffleSymbols(void); /* Shuffle symbol ids stored in precedence field */
void UnshuffleSymbols(void); /* Restore shuffled symbol ids */
int32_t ShuffleLiterals(cmprefix *); /* Shuffle literals and equality root terms in a clause */
cmprefix *SelectUnprocClause(void); /* Select and unlink an unprocessed clause */
